from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.utils import RFUniverseController
import numpy as np


class FrankaPushEnv(RFUniverseBaseEnv):

    def __init__(self, executable_file=None):
        super().__init__(
            executable_file,
            camera_channel=True,
            rigidbody_channel=True,
            articulation_channel=True
        )

        self.prev_joint_positions = np.array([0.0 for i in range(8)], dtype=float)
        self.ik_controller = RFUniverseController('franka')

    def step(self, a: np.ndarray):
        """
        Params:
            a: 4d numpy array. The first 3 dimensions are for the unity-position of Franka grasp point,
               while the 4th dimension is for gripper's width.
        """
        assert a.shape == (4,), \
            'The shape of action must be (4,), but got {}'.format(a.shape)

        eef_pos = a[0:3]
        joint_positions = self.ik_controller.calculate_ik(eef_pos)
        joint_positions.append(float(a[3]))
        velocities = [1.0 for i in range(8)]
        a = np.array(joint_positions + velocities)

        self._set_franka_joints(a)
        self._wait_for_moving()
        self.t += 1

        self._update_joint_positions()
        return self._get_obs()

    def _get_obs(self):
        # Read message from channels
        return self.rigidbody_channel.data

    def reset(self):
        self.t = 0
        self.r = 0
        self.env.reset()

        return self._get_obs()

    def seed(self):
        pass

    def _set_franka_joints(self, a: np.ndarray):
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=0,
            joint_positions=list(a[0:7]),
            speed_scales=list(a[8:15])
        )
        self._step()

        # Since a[7] is gripper's width, convert it to joint position value
        a[7] = -1 * a[7] / 2

        self.articulation_channel.set_action(
            'SetJointPosition',
            index=1,
            joint_positions=[a[7], a[7]],
            speed_scales=[a[15], a[15]]
        )
        self._step()

    def _wait_for_moving(self):
        self.articulation_channel.data.clear()
        while not 0 in self.articulation_channel.data.keys():
            self._step()
        while not (self.articulation_channel.data[0]['all_stable'] and self.articulation_channel.data[1]['all_stable']):
            self._step()

    def _update_joint_positions(self):
        data = self.articulation_channel.data
        arm_joint_positions = data[0]['joint_positions']
        gripper_joint_positions = data[1]['joint_positions']

        self.prev_joint_positions[0:7] = np.array(arm_joint_positions)
        self.prev_joint_positions[7] = abs(gripper_joint_positions[0]) + abs(gripper_joint_positions[1])

    def _calculate_speed(self, a: np.ndarray):
        relative_joint_positions = abs(a[0:7] - self.prev_joint_positions[0:7])
        speed = relative_joint_positions / relative_joint_positions.min()
        speed = np.concatenate(speed, np.array([1.0]))

        return speed



from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.envs import RFUniverseGymWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding
import math
import copy


class Robotiq85NailCanEnv(RFUniverseGymWrapper):
    def __init__(
            self,
            max_episode_length=40,
            prismatic_factor=0.05,
            revolute_facotr=0,
            min_success_degree=45,
            radius=0.15,
            executable_file=None,
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
        )
        self.max_episode_length = max_episode_length
        self.prismatic_factor = prismatic_factor
        self.revolute_factor = revolute_facotr
        self.bit_wise_factor = np.array([
            self.prismatic_factor,
            self.prismatic_factor,
            self.prismatic_factor,
            self.revolute_factor
        ])
        self.min_success_degree = min_success_degree
        self.radius = radius

        self.t = 0
        self.action_space = spaces.Box(
            low=-1, high=1, shape=(4,), dtype=np.float32
        )
        obs = self._get_obs()
        self.observation_space = spaces.Box(
            low=-np.inf, high=np.inf, shape=obs.shape, dtype=np.float32
        )

    def step(self, a: np.ndarray):
        action = a.copy()
        action_ctrl = action * self.bit_wise_factor
        curr_state = self._get_gripper_extra_param()
        target_state = action_ctrl + curr_state

        self._set_target_state(target_state)
        self.t += 1

        obs = self._get_obs()
        reward = self._compute_reward(obs)
        info = {'is_success': reward * 10 > self.min_success_degree}
        done = False
        if self.t == self.max_episode_length:
            done = True
            obs = self.reset()

        return obs, reward, done, info

    def reset(self):
        self.env.reset()
        self.t = 0

        theta = self._generate_random_float(-1 * math.pi / 2, math.pi / 2)
        self.articulation_channel.set_action(
            'SetJointPositionDirectly',
            index=0,
            joint_positions=[-self.radius * math.cos(theta), 0, -self.radius * math.sin(theta), 45]
        )
        self._step()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode='human'):
        self._step()

    def _get_obs(self):
        gripper_position = np.array(self.articulation_channel.data[0]['positions'][17])
        gripper_velocity = np.array(self.articulation_channel.data[0]['velocities'][17])
        gripper_extra_param = self._get_gripper_extra_param()

        # print(self.articulation_channel.data[1])
        ring_pos = np.array(self.articulation_channel.data[1]['positions'][1])
        ring_rotation = np.array([self.articulation_channel.data[1]['joint_positions'][0]])

        obs = np.concatenate((
            gripper_position, gripper_velocity, gripper_extra_param,
            ring_pos, ring_rotation
        ))

        return obs

    def _get_gripper_extra_param(self):
        joint_positions = np.array(self.articulation_channel.data[0]['joint_positions'])
        x_prismatic_joint_position = joint_positions[0]
        y_prismatic_joint_position = joint_positions[1]
        z_prismatic_joint_position = joint_positions[2]
        revolute_joint_position = joint_positions[3]

        return np.array(joint_positions)

    def _set_target_state(self, target_state):
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=0,
            joint_positions=list(target_state)
        )
        self._step()

    def _compute_reward(self, obs):
        ring_rotation = obs[-1] * -1
        if ring_rotation < 5:
            return 0
        else:
            return ring_rotation / 10

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, \
            'Min value is {}, while max value is {}.'.format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float


from pyrfuniverse.envs import RFUniverseGymWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding


class RollerEnv(RFUniverseGymWrapper):
    metadata = {'render.modes': ['human']}

    def __init__(self, executable_file=None):
        super().__init__(
            executable_file,
            rigidbody_channel=True,
            game_object_channel=True
        )
        self.action_space = spaces.Discrete(5)
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(6,), dtype=np.float32)
        self.t = 0
        self.max_steps = 100
        self.strength = 250

    def step(self, a):
        direction = None
        if a == 0:
            direction = [0, 0, 0]
        elif a == 1:
            direction = [1, 0, 0]
        elif a == 2:
            direction = [-1, 0, 0]
        elif a == 3:
            direction = [0, 0, 1]
        elif a == 4:
            direction = [0, 0, -1]
        direction = np.array(direction)
        force = direction * self.strength

        self.rigidbody_channel.set_action(
            'AddForce',
            index=0,
            force=list(force)
        )
        self._step()

        done = False
        success = self._check_success()
        info = {'is_success': False}
        reward = -1 / self.max_steps
        self.t += 1

        if success:
            reward = 1
            done = True
            info['is_success'] = True
        elif self._check_fail():
            reward = -1
            done = True
        elif self.t == self.max_steps:
            done = True

        if done:
            self.reset()

        return self._get_obs(), reward, done, info

    def _check_fail(self):
        roller_position = self.rigidbody_channel.data[0]['position']

        if roller_position[1] < -1:
            return True
        elif roller_position[0] < -19 or roller_position[0] > 19 or roller_position[2] < -19 \
            or roller_position[2] > 19:
            return True
        else:
            return False

    def _check_success(self):
        roller_position = self.rigidbody_channel.data[0]['position']
        roller_position_2d = np.array([roller_position[0], roller_position[2]])
        target_position = self.game_object_channel.data[0]['position']
        target_position_2d = np.array([target_position[0], target_position[2]])

        distance = abs(target_position_2d - roller_position_2d)
        if distance[0] < 2.5 and distance[1] < 2.5:
            return True
        else:
            return False

    def _get_obs(self):
        roller_position = self.rigidbody_channel.data[0]['position']
        roller_position_2d = np.array([roller_position[0], roller_position[2]])
        roller_velocity = self.rigidbody_channel.data[0]['velocity']
        roller_velocity_2d = np.array([roller_velocity[0], roller_velocity[2]])
        target_position = self.game_object_channel.data[0]['position']
        target_position_2d = np.array([target_position[0], target_position[2]])

        return np.concatenate((roller_position_2d, roller_velocity_2d, target_position_2d))

    def reset(self):
        self.t = 0

        roller_pos = self.np_random.uniform(-15, 15, size=2)
        target_pos = roller_pos.copy()
        while self._compute_distance(roller_pos, target_pos) < 5:
            target_pos = self.np_random.uniform(-15, 15, size=2)

        self.env_param_channel.set_float_parameter('rollerPosX', roller_pos[0])
        self.env_param_channel.set_float_parameter('rollerPosZ', roller_pos[1])
        self.env_param_channel.set_float_parameter('targetPosX', target_pos[0])
        self.env_param_channel.set_float_parameter('targetPosZ', target_pos[1])

        self.env.reset()
        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, \
            'Min value is {}, while max value is {}.'.format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _compute_distance(self, goal_a, goal_b):
        assert goal_a.shape == goal_b.shape
        return np.linalg.norm(goal_a - goal_b, axis=-1)


class RollerEnvV0(RollerEnv):

    def __init__(self):
        super().__init__('/home/haoyuan/workspace/rfuniverse/build/Roller/RFUniverse.x86_64')



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import numpy as np
import math


def average_interpolate_with_max_step_length(start: np.ndarray, terminal: np.ndarray, max_step_length):
    assert start.shape == terminal.shape
    distance = terminal - start
    num_steps = int(abs(distance).max() / max_step_length) + 1
    unit = distance / num_steps

    intermediate_nodes = []
    for i in range(num_steps):
        intermediate_nodes.append(start + unit * (i + 1))
    intermediate_nodes.append(terminal)

    return np.array(intermediate_nodes)


def average_interpolate(start: np.ndarray, terminal: np.ndarray, num_steps):
    assert start.shape == terminal.shape
    distance = terminal - start
    unit = distance / num_steps

    intermediate_nodes = []
    for i in range(num_steps):
        intermediate_nodes.append(start + unit * (i + 1))

    return np.array(intermediate_nodes)


def sine_interpolate(start: np.ndarray, terminal: np.ndarray, num_steps):
    assert start.shape == terminal.shape
    distance = terminal - start

    intermediate_nodes = []
    for i in range(num_steps):
        step = distance / 2 * (math.sin(math.pi / num_steps * (i + 1) - math.pi / 2) + 1)
        node = start + step
        intermediate_nodes.append(node)

    return np.array(intermediate_nodes)


def rotate_by_y_axis_interpolate(start: np.array, center: np.array, moving_degree: float, num_steps: int):
    """
    The rotation is default to be anti-clockwise.
    Args:
        start: The start position in Unity.
        center: The center position in Unity.
        moving_degree: The amount of rotation need to rotate, in degree.
        num_steps: The number of steps of this movement.

    Returns:
        The interpolated positions of this movement, in `np.ndarray` format
    """
    start_2d = np.array([start[0], start[2]])
    center_2d = np.array([center[0], center[2]])
    relative_2d = start_2d - center_2d
    radius = np.linalg.norm(relative_2d)
    init_radius = math.acos(float(relative_2d[0]) / radius)
    if float(relative_2d[1]) < 0:
        init_radius = -1 * init_radius
    delta_radius = moving_degree * math.pi / 180 / num_steps
    intermediate_nodes = []
    for i in range(num_steps):
        curr_target_radius = init_radius + delta_radius * (i + 1)
        relative_x = radius * math.cos(curr_target_radius)
        relative_z = radius * math.sin(curr_target_radius)
        intermediate_node = np.array([center_2d[0] + relative_x, start[1], center_2d[1] + relative_z])
        intermediate_nodes.append(intermediate_node)

    return np.array(intermediate_nodes)


def joint_positions_interpolation(env: RFUniverseBaseEnv, body_id, target_joint_positions, max_step_degree=3):
    curr_jp = np.array(env.articulation_channel.data[body_id]['joint_positions'])
    target_jp = np.array(target_joint_positions)
    return average_interpolate(curr_jp, target_jp, max_step_degree)


def pos_interpolation(env: RFUniverseBaseEnv, body_id, part_index, target_pos, max_step_length=0.05):
    curr_pos = np.array(env.articulation_channel.data[body_id]['positions'][part_index])
    targ_pos = np.array(target_pos)
    return average_interpolate(curr_pos, targ_pos, max_step_length)





from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr
import cv2
import numpy as np

env = RFUniverseBaseEnv(assets=['Camera', 'GameObject_Box'])

camera = env.InstanceObject(name='Camera', id=123456, attr_type=attr.CameraAttr)
box = env.InstanceObject(name='GameObject_Box', attr_type=attr.GameObjectAttr)
box.SetTransform(position=[0, 0.05, 0.5], scale=[0.1, 0.1, 0.1])
box.SetColor([1, 0, 0, 1])
env.step()
camera.SetTransform(position=[0, 0.25, 0], rotation=[30, 0, 0])
camera.LookAt(target=box.data['position'])
camera.GetDepth(width=512, height=512, zero_dis=1, one_dis=5)
camera.GetDepthEXR(width=512, height=512)
camera.GetRGB(width=512, height=512)
env.step()
print(camera.data['depth'])
print(camera.data['depth_exr'])
print(camera.data['rgb'])
image = np.frombuffer(camera.data['rgb'], dtype=np.uint8)
image = cv2.imdecode(image, cv2.IMREAD_COLOR)
print(image.shape)
env.close()
cv2.imshow("rgb", image)
cv2.waitKey(0)



import pyrfuniverse.attributes as attr


class HumanbodyAttr(attr.BaseAttr):
    """
    Human body Inverse Kinematic class.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['move_done']: Whether the movement has finished.

            self.data['rotate_done']: Whether the rotation has finished.
        """
        super().parse_message(data)

    def HumanIKTargetDoMove(self, index: int, position: list, duration: float, speed_based: bool = True,
                            relative: bool = False):
        """
        Human body Inverse Kinematics target movement.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
            position: A list of length 3, representing the position.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `position` is relative; otherwise, `position` is absolute.
        """
        if position is not None:
            assert len(position) == 3, 'position length must be 3'
            position = [float(i) for i in position]
        self._send_data('HumanIKTargetDoMove', index, position, float(duration), speed_based, relative)

    def HumanIKTargetDoRotate(self, index: int, rotation: list, duration: float, speed_based: bool = True,
                              relative: bool = False):
        """
        Human body Inverse Kinematics target rotation.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
            rotation: A list of length 3, representing the rotation.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `rotation` is relative; otherwise, `rotation` is absolute.
        """
        if rotation is not None:
            assert len(rotation) == 3, 'rotation length must be 3'
            rotation = [float(i) for i in rotation]
        self._send_data('HumanIKTargetDoRotate', index, rotation, float(duration), speed_based, relative)

    def HumanIKTargetDoRotateQuaternion(self, index: int, quaternion: list, duration: float, speed_based: bool = True,
                                        relative: bool = False):
        """
        Human body Inverse Kinematics target rotation using quaternion.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
            quaternion: A list of length 4, representing the quaternion.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `quaternion` is relative; otherwise, `quaternion` is absolute.
        """
        if quaternion is not None:
            assert len(quaternion) == 4, 'quaternion length must be 4'
            quaternion = [float(i) for i in quaternion]
        self._send_data('HumanIKTargetDoRotateQuaternion', index, quaternion, float(duration), speed_based, relative)

    def HumanIKTargetDoComplete(self, index: int):
        """
        Make the human body IK target movement / rotation complete directly.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
        """
        self._send_data('HumanIKTargetDoComplete', index)

    def HumanIKTargetDoKill(self, index: int):
        """
        Make the human body IK target movement / rotation stop.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
        """
        self._send_data('HumanIKTargetDoKill', index)

    def WaitDo(self):
        """
        Wait for the human body IK target movement / rotation complete.
        """
        while not self.data['move_done'] or not self.data['rotate_done']:
            self.env.step()




from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.side_channel import SideChannel
from pyrfuniverse.side_channel import IncomingMessage
from pyrfuniverse.envs import RFUniverseGymWrapper

import numpy as np
from gym import spaces
from gym.utils import seeding
import uuid


class CollisionAndCleaningDetectionChannel(SideChannel):

    def __init__(self, agents, num_floors=64):
        super().__init__(uuid.UUID('bee25cbc-09e2-11ec-9e67-18c04d443e7d'))
        self._agent_collisions = {}
        self._agent_cleanings = {}
        self._num_collisions = 0
        self._cleaning_states = np.zeros((num_floors,), dtype=bool)
        for agent in agents:
            self._agent_collisions[agent] = 0
            self._agent_cleanings[agent] = 0

    def on_message_received(self, msg: IncomingMessage) -> None:
        self._num_collisions += msg.read_int32()
        self._cleaning_states = np.array([msg.read_float32_list()]) > 0
        for i in range(len(self._agent_collisions.items())):
            agent_name = msg.read_string()
            num_collisions = msg.read_int32()
            num_cleanings = msg.read_int32()

            if agent_name == '':
                continue
            if agent_name[-7:] == '(Clone)':
                agent_name = agent_name[:-7]
            self._agent_collisions[agent_name] += num_collisions
            self._agent_cleanings[agent_name] += num_cleanings

    def get_agent_collisions(self):
        agent_collisions = self._agent_collisions.copy()
        for key, value in self._agent_collisions.items():
            self._agent_collisions[key] = 0

        return agent_collisions

    def get_agent_cleanings(self):
        agent_cleanings = self._agent_cleanings.copy()
        for key, value in self._agent_cleanings.items():
            self._agent_cleanings[key] = 0

        return agent_cleanings

    def get_num_collisions(self):
        num_collisions = self._num_collisions
        self._num_collisions = 0

        return num_collisions

    def get_cleaning_states(self):
        return self._cleaning_states.copy()

    def clear(self):
        self._num_collisions = 0
        for key, value in self._agent_collisions.items():
            self._agent_collisions[key] = 0
            self._agent_cleanings[key] = 0
        self._cleaning_states = np.zeros(self._cleaning_states.shape, dtype=bool)


class CleanerEnv(RFUniverseGymWrapper):
    metadata = {'render.modes': ['human']}

    def __init__(
            self,
            asset_bundle_file,
            max_episode_length=100,
            reset_on_collision=True,
            velocity_reward=False,
            obs_type='multi',
            grid_reward_per_step=False,
            num_agents=3,
            strength=1000,
            collision_multiplier=5,
            num_floors=64,
            executable_file=None,
    ):
        self.num_agents = num_agents
        self.num_floors = num_floors
        self.agent_names = ['agent_' + str(i) for i in range(self.num_agents)]
        self.channel = CollisionAndCleaningDetectionChannel(self.agent_names, self.num_floors)
        super().__init__(
            executable_file=executable_file,
            custom_channels=[self.channel],
            rigidbody_channel=True,
        )

        self.asset_bundle_file = asset_bundle_file
        self.max_episode_length = max_episode_length
        self.reset_on_collision = reset_on_collision
        self.strength = strength
        self.collision_multiplier = collision_multiplier
        self.velocity_reward = velocity_reward
        self.obs_type = obs_type
        self.grid_reward_per_step = grid_reward_per_step

        self.agent_name = 'CleanRobot'
        self.y_offset = 0.1
        self.reset_agent_min_distance = 1.0
        self.world_range_low = np.array([-3, self.y_offset, -3])
        self.world_range_high = np.array([3, self.y_offset, 3])
        self.seed()

        self.action_space = spaces.Box(low=-1, high=1, shape=(2 * self.num_agents,), dtype=np.float32)

        if self.obs_type == 'multi':
            self.observation_space = spaces.Dict({
                'agent_obs': spaces.Box(low=-np.inf, high=np.inf, shape=(4 * self.num_agents,), dtype=float),
                'floor_obs': spaces.MultiBinary(self.num_floors)
            })
        else:
            self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(4 * self.num_agents,), dtype=float)
        self._env_setup()
        self.t = 0
        self.num_clean_grids = 0

    def step(self, action: np.ndarray):
        a = action.copy()

        for i in range(self.num_agents):
            target_force = np.array([a[2 * i], 0, a[2 * i + 1]]) * self.strength
            self._set_agent_force(i, target_force)

        done = False
        num_collisions = self.channel.get_num_collisions()
        cleaning_states = self.channel.get_cleaning_states().astype(np.int)
        total_num_clean_grids = np.sum(cleaning_states)

        reward = 0
        if self.grid_reward_per_step:
            assert total_num_clean_grids >= self.num_clean_grids
            reward += total_num_clean_grids - self.num_clean_grids
        else:
            reward += total_num_clean_grids / 10
        if self.velocity_reward:
            reward += self._get_total_velocity() / self.num_agents
        if not self.reset_on_collision:
            reward += -1 * num_collisions * self.collision_multiplier

        info = {'is_success': total_num_clean_grids > self.num_floors * 0.8}
        self.t += 1
        self.num_clean_grids = total_num_clean_grids

        if self.reset_on_collision and num_collisions > 0:
            reward = -1 * self.collision_multiplier * num_collisions
            self.reset()
            done = True

        elif self.t == self.max_episode_length:
            self.reset()
            done = True

        return self._get_obs(), float(reward), done, info

    def reset(self):
        positions = []
        for i in range(self.num_agents):
            while True:
                agent_position = self.np_random.uniform(self.world_range_low, self.world_range_high)
                if self._check_reset_position_legality(positions, agent_position):
                    break
            positions.append(agent_position.copy())
            self._set_agent_position(i, agent_position)

        self.t = 0
        self.num_clean_grids = 0
        self.env.reset()
        self.channel.clear()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode='human'):
        self._step()

    def _env_setup(self):
        for i in range(self.num_agents):
            agent_position = self.np_random.uniform(self.world_range_low, self.world_range_high)
            self.asset_channel.set_action(
                'LoadRigidbodyWithName',
                filename=self.asset_bundle_file,
                name=self.agent_name + ' ' + str(i),
                replace_name=self.agent_names[i],
                position=list(agent_position)
            )
            self._step()

    def _get_obs(self):
        agent_obs = np.array([])
        for i in range(self.num_agents):
            agent_position = self.rigidbody_channel.data[i]['position']
            agent_velocity = self.rigidbody_channel.data[i]['velocity']
            # We ignore y-axis
            this_agent_obs = np.array([agent_position[0], agent_position[2], agent_velocity[0], agent_velocity[2]])
            agent_obs = np.concatenate((agent_obs, this_agent_obs))

        floor_obs = self.channel.get_cleaning_states().astype(np.int8)

        if self.obs_type == 'multi':
            return {
                'agent_obs': agent_obs.copy(),
                'floor_obs': floor_obs.copy()[0]
            }
        else:
            return agent_obs.copy()

    def _set_agent_position(self, index, position):
        self.rigidbody_channel.set_action(
            'SetTransform',
            index=index,
            position=list(position),
            rotation=[0, 0, 0]
        )
        self._step()

    def _set_agent_force(self, index, force):
        self.rigidbody_channel.set_action(
            'AddForce',
            index=index,
            force=force
        )
        self._step()

    def _get_total_velocity(self):
        total_velocity = 0
        for i in range(self.num_agents):
            velocity = self.rigidbody_channel.data[i]['velocity']
            total_velocity += np.linalg.norm(np.array([velocity[0], velocity[2]]), axis=-1)

        return total_velocity

    def _check_reset_position_legality(self, positions, agent_pos):
        for position in positions:
            if np.linalg.norm(position - agent_pos) < self.reset_agent_min_distance:
                return False

        return True



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

env = RFUniverseBaseEnv(assets=['Collider_Box', 'Rigidbody_Sphere'])

box1 = env.InstanceObject(name='Collider_Box', attr_type=attr.ColliderAttr)
box1.SetTransform(position=[-0.5, 0.5, 0], scale=[0.1, 1, 1])
box2 = env.InstanceObject(name='Collider_Box', attr_type=attr.ColliderAttr)
box2.SetTransform(position=[0.5, 0.5, 0], scale=[0.1, 1, 1])
box3 = env.InstanceObject(name='Collider_Box', attr_type=attr.ColliderAttr)
box3.SetTransform(position=[0, 0.5, 0.5], scale=[1, 1, 0.1])
box4 = env.InstanceObject(name='Collider_Box', attr_type=attr.ColliderAttr)
box4.SetTransform(position=[0, 0.5, -0.5], scale=[1, 1, 0.1])
sphere = env.InstanceObject(name='Rigidbody_Sphere', attr_type=attr.RigidbodyAttr)
sphere.SetTransform(position=[0, 0.5, 0], scale=[0.5, 0.5, 0.5])
env.Pend()

env.SaveScene('test_scene.json')
env.ClearScene()
env.Pend()

env.LoadSceneAsync('test_scene.json')
env.Pend()
env.close()



import base64
import cv2
import numpy as np
import pyrfuniverse.attributes as attr
import pyrfuniverse.utils.active_depth_generate as active_depth


class ActiveLightSensorAttr(attr.CameraAttr):
    """
    Class of IR-based depth sensor, which can simulate the noise of 
    real-world depth camera and produce depth image in similar pattern.
    """
    def __init__(self, env, id: int, data=None):
        super().__init__(env, id, data)
        self.main_intrinsic_matrix = np.eye(3)
        self.ir_intrinsic_matrix = np.eye(3)

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.\n
            self.data['active_depth']: IR-based depth, shape = (w,h)
        """
        super().parse_message(data)
        if 'ir_left' in self.data and 'ir_right' in self.data:
            self.data['ir_left'] = base64.b64decode(self.data['ir_left'])
            self.data['ir_right'] = base64.b64decode(self.data['ir_right'])

            image_left = np.frombuffer(self.data['ir_left'], dtype=np.uint8)
            image_left = cv2.imdecode(image_left, cv2.IMREAD_COLOR)[..., 2]
            image_right = np.frombuffer(self.data['ir_right'], dtype=np.uint8)
            image_right = cv2.imdecode(image_right, cv2.IMREAD_COLOR)[..., 2]
            left_extrinsic_matrix = np.array(
                [[0., -1., 0., -0.0175], [0., 0., -1., 0.], [1., 0., 0., 0.], [0., 0., 0., 1.]])
            right_extrinsic_matrix = np.array(
                [[0., -1., 0., -0.072], [0., 0., -1., 0.], [1., 0., 0., 0.], [0., 0., 0., 1.]])
            main_extrinsic_matrix = np.array([[0., -1., 0., 0.], [0., 0., -1., 0.], [1., 0., 0., 0.], [0., 0., 0., 1.]])
            self.data['active_depth'] = active_depth.calc_main_depth_from_left_right_ir(image_left, image_right,
                                                                                        left_extrinsic_matrix,
                                                                                        right_extrinsic_matrix,
                                                                                        main_extrinsic_matrix,
                                                                                        self.ir_intrinsic_matrix,
                                                                                        self.ir_intrinsic_matrix,
                                                                                        self.main_intrinsic_matrix,
                                                                                        lr_consistency=False,
                                                                                        main_cam_size=(
                                                                                            self.main_intrinsic_matrix[
                                                                                                0, 2] * 2,
                                                                                            self.main_intrinsic_matrix[
                                                                                                1, 2] * 2),
                                                                                        ndisp=128,
                                                                                        use_census=True,
                                                                                        register_depth=True,
                                                                                        census_wsize=7,
                                                                                        use_noise=False)
            self.data['active_depth'][self.data['active_depth'] > 8.] = 0
            self.data['active_depth'][self.data['active_depth'] < 0.1] = 0

    def GetActiveDepth(self, main_intrinsic_matrix_local: np.ndarray, ir_intrinsic_matrix_local: np.ndarray):
        """
        Get IR-based depth image.

        Args:
            main_intrinsic_matrix_local: np.ndarray The intrinsic matrix of main camera.
            ir_intrinsic_matrix_local: np.ndarray The intrinsic matrix of IR-based camera.
        """
        self.main_intrinsic_matrix = main_intrinsic_matrix_local
        self.ir_intrinsic_matrix = ir_intrinsic_matrix_local
        self._send_data('GetActiveDepth', self.ir_intrinsic_matrix)



import cv2
import numpy as np
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

env = RFUniverseBaseEnv(scene_file='SimpleYCBModel.json')
for i in env.attrs:
    print(i)
    print(env.attrs[i])

env.step(100)
camera = env.GetAttr(981613)
camera.GetRGB(512, 512)
camera.GetID(512, 512)
camera.GetNormal(512, 512)
camera.GetDepth(0.1, 2., 512, 512, )
camera.GetAmodalMask(655797, 512, 512)
camera.Get2DBBox(512, 512)
camera.Get3DBBox()

env.step()

#rgb
rgb = np.frombuffer(camera.data['rgb'], dtype=np.uint8)
rgb = cv2.imdecode(rgb, cv2.IMREAD_COLOR)
cv2.imshow("rgb", rgb)
cv2.waitKey(0)

#normal
normal = np.frombuffer(camera.data['normal'], dtype=np.uint8)
normal = cv2.imdecode(normal, cv2.IMREAD_COLOR)
cv2.imshow("normal", normal)
cv2.waitKey(0)

#depth
depth = np.frombuffer(camera.data['depth'], dtype=np.uint8)
depth = cv2.imdecode(depth, cv2.IMREAD_COLOR)
cv2.imshow("depth", depth)
cv2.waitKey(0)

#amodal_mask
amodal_mask = np.frombuffer(camera.data['amodal_mask'], dtype=np.uint8)
amodal_mask = cv2.imdecode(amodal_mask, cv2.IMREAD_COLOR)
cv2.imshow("amodal_mask", amodal_mask)
cv2.waitKey(0)

#2d_bounding_box
id_map = np.frombuffer(camera.data['id_map'], dtype=np.uint8)
id_map = cv2.imdecode(id_map, cv2.IMREAD_COLOR)

print('2d_bounding_box:')
for i in camera.data['2d_bounding_box']:
    print(i)
    print(camera.data['2d_bounding_box'][i])
    center = camera.data['2d_bounding_box'][i][0:2]
    size = camera.data['2d_bounding_box'][i][2:4]
    tl_point = (int(center[0] + size[0] / 2), int(512 - center[1] + size[1] / 2))
    br_point = (int(center[0] - size[0] / 2), int(512 - center[1] - size[1] / 2))
    cv2.rectangle(id_map, tl_point, br_point, (255, 255, 255), 1)

#3d_bounding_box
print('3d_bounding_box:')
for i in camera.data['3d_bounding_box']:
    print(i)
    print(camera.data['3d_bounding_box'][i])

cv2.imshow("id_map", id_map)
cv2.waitKey(0)

env.Pend()
env.close()




# from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
# try:
#     from pyrfuniverse.envs.gym_wrapper_env import RFUniverseGymWrapper
#     from pyrfuniverse.envs.gym_goal_wrapper import RFUniverseGymGoalWrapper
# except ImportError:
#     pass
# from pyrfuniverse.envs.franka_grasp_env import FrankaGraspEnv
# from pyrfuniverse.envs.franka_push_env import FrankaPushEnv
# from pyrfuniverse.envs.balance_ball_env import BalanceBallEnv
# from pyrfuniverse.envs.balance_ball_env import BalanceBallEnvV0
# from pyrfuniverse.envs.bouncer_env import BouncerEnv
# from pyrfuniverse.envs.bouncer_env import BouncerEnvV0
# from pyrfuniverse.envs.roller_env import RollerEnv
# from pyrfuniverse.envs.roller_env import RollerEnvV0
# from pyrfuniverse.envs.gripper_nail import NailCardEnv
# from pyrfuniverse.envs.gripper_nail import NailCanEnv
# from pyrfuniverse.envs.multi_agent_navigation_env import MultiAgentNavigationEnv
# from pyrfuniverse.envs.tobor_robotiq85_manipulation_env import ToborRobotiq85ManipulationEnv
# from pyrfuniverse.envs.ur5_box_env import Ur5BoxEnv
# from pyrfuniverse.envs.ur5_drawer_env import Ur5DrawerEnv

# Other projects based on RFUniverse
# import pyrfuniverse.envs.robotube
# import pyrfuniverse.envs.rcareworld

# __all__ = [
#     'RFUniverseBaseEnv', 'FrankaGraspEnv', 'FrankaPushEnv', 'BalanceBallEnv',
#     'RFUniverseGymWrapper', 'BalanceBallEnvV0', 'RFUniverseGymGoalWrapper',
#     'BouncerEnv', 'BouncerEnvV0', 'RollerEnv', 'RollerEnvV0', 'NailCardEnv',
#     'MultiAgentNavigationEnv', 'ToborRobotiq85ManipulationEnv', 'Ur5BoxEnv',
#     'Ur5DrawerEnv',
# ]
#
#
# rfuniverse_build_base_root = '/home/haoyuan/workspace/rfuniverse/rfuniverse/rfuniverse_build/'
#
# try:
#     from gym.envs.registration import register
#
#     register(
#         id='BalanceBallEnv-v0',
#         entry_point='pyrfuniverse.envs:BalanceBallEnvV0'
#     )
#
#     register(
#         id='BouncerEnv-v0',
#         entry_point='pyrfuniverse.envs:BouncerEnvV0'
#     )
#
#     register(
#         id='RollerEnv-v0',
#         entry_point='pyrfuniverse.envs:RollerEnvV0'
#     )
#
#     register(
#         id='MultiAgentNavigation-v1',
#         entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'num_agents': 5,
#         },
#     )
#
#     register(
#         id='MultiAgentNavigation-v2',
#         entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'num_agents': 5,
#             'reset_on_collision': True
#         },
#     )
#
#     register(
#         id='MultiAgentNavigation-v3',
#         entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'num_agents': 3,
#         },
#     )
#
#     register(
#         id='MultiAgentNavigation-v4',
#         entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'num_agents': 3,
#             'reset_on_collision': True
#         },
#     )
#
#     # Franka robotics
#     for reward_type in ['sparse', 'dense']:
#         suffix = 'Dense' if reward_type == 'dense' else ''
#         kwargs = {
#             'executable_file': rfuniverse_build_base_root + 'FrankaRoboticsServer/RFUniverse.x86_64',
#             'max_episode_length': 50,
#             'reward_type': reward_type,
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody'
#         }
#
#         register(
#             id='FrankaPickAndPlace{}-v1'.format(suffix),
#             entry_point='pyrfuniverse.envs.robotics:FrankaPickAndPlaceEnv',
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )
#
#         register(
#             id='FrankaReach{}-v1'.format(suffix),
#             entry_point='pyrfuniverse.envs.robotics:FrankaReachEnv',
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )
#
#         register(
#             id='FrankaPush{}-v1'.format(suffix),
#             entry_point='pyrfuniverse.envs.robotics:FrankaPushEnv',
#             kwargs=kwargs,
#             max_episode_steps=50,
#         )
#
#     # Franka-Cloth env
#     franka_cloth_kwargs = {
#         'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/obi_cloth',
#         'executable_file': rfuniverse_build_base_root + 'FrankaCloth/RFUniverse.x86_64',
#         'reward_type': 'dense'
#     }
#     register(
#         id='FrankaCloth-v1',
#         entry_point='pyrfuniverse.envs.robotics:FrankaClothEnv',
#         kwargs=franka_cloth_kwargs,
#         max_episode_steps=50
#     )
#
#     # Franka-Softbody env
#     franka_softbody_kwargs = {
#         'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/softbody',
#         # 'executable_file': rfuniverse_build_base_root + 'FrankaSoftbodyServer/RFUniverse.x86_64',
#         'executable_file': None,
#         'reward_type': 'sparse'
#     }
#     register(
#         id='FrankaSoftbody-v1',
#         entry_point='pyrfuniverse.envs.robotics:FrankaSoftbodyEnv',
#         kwargs=franka_softbody_kwargs,
#         max_episode_steps=50
#     )
#
#     # Franka-ClothFold env
#     franka_cloth_fold_kwargs = {
#         'asset_bundle_file': '/home/haoyuan/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/obi_cloth',
#         # 'executable_file': rfuniverse_build_base_root + 'FrankaClothFoldServer/RFUniverse.x86_64',
#         'executable_file': None,
#         'reward_type': 'sparse'
#     }
#     register(
#         id='FrankaClothFold-v1',
#         entry_point='pyrfuniverse.envs.robotics:FrankaClothFoldEnv',
#         kwargs=franka_cloth_fold_kwargs,
#         max_episode_steps=20
#     )
#
#     # A relatively easy and naive NailCard environment
#     register(
#         id='NailCard-v1',
#         entry_point='pyrfuniverse.envs:NailCardEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'NailCardServer/RFUniverse.x86_64',
#             'rotation_factor': 0,
#             'goal_baseline': 0.02,
#         },
#         max_episode_steps=50,
#     )
#
#     # The normal version
#     register(
#         id='NailCard-v2',
#         entry_point='pyrfuniverse.envs:NailCardEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'NailCardServer/RFUniverse.x86_64',
#         },
#         max_episode_steps=50,
#     )
#
#     # Fixed the nail movement
#     register(
#         id='NailCard-v3',
#         entry_point='pyrfuniverse.envs:NailCardEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'NailCardServer/RFUniverse.x86_64',
#             'nail_movement_factor': 0
#         },
#         max_episode_steps=50,
#     )
#
#     # A relatively easy and naive NailCard environment
#     register(
#         id='Robotiq85NailCard-v1',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCardEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCardServer/RFUniverse.x86_64',
#             'rotation_factor': 0,
#             'goal_baseline': 0.02,
#         },
#         max_episode_steps=50,
#     )
#
#     # The normal version
#     register(
#         id='Robotiq85NailCard-v2',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCardEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCardServer/RFUniverse.x86_64',
#         },
#         max_episode_steps=50,
#     )
#
#     # Fixed the nail movement
#     register(
#         id='Robotiq85NailCard-v3',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCardEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCardServer/RFUniverse.x86_64',
#             'vertical_movement_factor': 0.02,
#             'nail_movement_factor': 0
#         },
#         max_episode_steps=50,
#     )
#
#     # Robotiq85 Nail Coin
#     register(
#         id='Robotiq85NailCoin-v1',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCoinEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCoinServer/RFUniverse.x86_64',
#             'vertical_movement_factor': 0.02,
#             'nail_movement_factor': 0
#         },
#         max_episode_steps=50,
#     )
#
#     # Nail-Can environment
#     register(
#         id='NailCan-v1',
#         entry_point='pyrfuniverse.envs:NailCanEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'NailCanServer/RFUniverse.x86_64',
#         }
#     )
#
#     register(
#         id='Robotiq85NailCan-v1',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCanEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCanServer/RFUniverse.x86_64',
#         }
#     )
#
#     register(
#         id='Robotiq85NailBook-v1',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailBookEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85NailBook/RFUniverse.x86_64',
#         }
#     )
#
#     # Robotiq85 insert
#     register(
#         id='Robotiq85Insert-v1',
#         entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85InsertEnv',
#         kwargs={
#             'executable_file': rfuniverse_build_base_root + 'Robotiq85InsertServer/RFUniverse.x86_64',
#         }
#     )
#
#     register(
#         id='Cleaner-v1',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerServer/RFUniverse.x86_64'
#         }
#     )
#
#     register(
#         id='Cleaner-v2',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'reset_on_collision': False,
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerServer/RFUniverse.x86_64',
#         }
#     )
#
#     register(
#         id='Cleaner-v3',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64'
#         }
#     )
#
#     register(
#         id='Cleaner-v4',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'reset_on_collision': False,
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64'
#         }
#     )
#
#     register(
#         id='Cleaner-v5',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64',
#             'velocity_reward': True
#         }
#     )
#
#     register(
#         id='Cleaner-v6',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64',
#             'velocity_reward': True,
#             'collision_multiplier': 50,
#         }
#     )
#
#     register(
#         id='Cleaner-v7',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'asset_bundle_file': '/home/haoyuan/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/rigidbody',
#             # 'executable_file': rfuniverse_build_base_root + 'CleanerWithWall/RFUniverse.x86_64',
#             'executable_file': None,
#             'velocity_reward': True,
#             'collision_multiplier': 50,
#             'obs_type': 'box',
#             'max_episode_length': 200
#         }
#     )
#
#     register(
#         id='Cleaner-v8',
#         entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
#         kwargs={
#             'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
#             'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64',
#             'grid_reward_per_step': True
#         }
#     )
#
#     register(
#         id='UR5Box-v1',
#         entry_point='pyrfuniverse.envs:Ur5BoxEnv',
#         kwargs={
#             'max_steps': 50,
#             'reward_type': 'sparse',
#             'min_open_angle': 30,
#             'executable_file': rfuniverse_build_base_root + 'UR5Box/RFUniverse.x86_64'
#         },
#         max_episode_steps=50,
#     )
#
#     register(
#         id='ToborPull-v1',
#         entry_point='pyrfuniverse.envs.tobor_robotics:ToborPushPullEnv',
#         kwargs={
#             'max_steps': 100,
#             'asset_bundle_file': '/home/haoyuan/workspace/rfuniverse/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/articulation',
#             'pull': True,
#             'executable_file': rfuniverse_build_base_root + 'ToborPushPull/RFUniverse.x86_64'
#         },
#         max_episode_steps=100
#     )
#
#     register(
#         id='ToborPush-v1',
#         entry_point='pyrfuniverse.envs.tobor_robotics:ToborPushPullEnv',
#         kwargs={
#             'max_steps': 100,
#             'asset_bundle_file': '/home/haoyuan/workspace/rfuniverse/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/articulation',
#             'pull': False,
#             'executable_file': rfuniverse_build_base_root + 'ToborPushPull/RFUniverse.x86_64'
#         },
#         max_episode_steps=100
#     )
#
# except ImportError:
#     print('No gym installed. Please install gym!')
#     pass



import random
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

env = RFUniverseBaseEnv(assets=['franka_panda'])

robot = env.InstanceObject(name='franka_panda', id=123456, attr_type=attr.ControllerAttr)
robot.SetIKTargetOffset(position=[0, 0.105, 0])
env.step()
gripper = env.GetAttr(1234560)
gripper.GripperOpen()
robot.IKTargetDoMove(position=[0, 0.5, 0.5], duration=0, speed_based=False)
robot.IKTargetDoRotate(rotation=[0, 45, 180], duration=0, speed_based=False)
env.step()
robot.WaitDo()

while 1:
    box1 = env.InstanceObject(name='Rigidbody_Box', id=111111, attr_type=attr.RigidbodyAttr)
    box1.SetTransform(position=[random.uniform(-0.5, -0.3), 0.03, random.uniform(0.3, 0.5)], scale=[0.06, 0.06, 0.06])
    box2 = env.InstanceObject(name='Rigidbody_Box', id=222222, attr_type=attr.RigidbodyAttr)
    box2.SetTransform(position=[random.uniform(0.3, 0.5), 0.03, random.uniform(0.3, 0.5)], scale=[0.06, 0.06, 0.06])
    env.step(100)

    position1 = box1.data['position']
    position2 = box2.data['position']

    robot.IKTargetDoMove(position=[position1[0], position1[1] + 0.5, position1[2]], duration=2, speed_based=False)
    env.step()
    robot.WaitDo()
    robot.IKTargetDoMove(position=[position1[0], position1[1], position1[2]], duration=2, speed_based=False)
    env.step()
    robot.WaitDo()
    gripper.GripperClose()
    env.step(50)
    robot.IKTargetDoMove(position=[0, 0.5, 0], duration=2, speed_based=False, relative=True)
    env.step()
    robot.WaitDo()
    robot.IKTargetDoMove(position=[position2[0], position2[1] + 0.5, position2[2]], duration=4, speed_based=False)
    env.step()
    robot.WaitDo()
    robot.IKTargetDoMove(position=[position2[0], position2[1] + 0.06, position2[2]], duration=2, speed_based=False)
    env.step()
    robot.WaitDo()
    gripper.GripperOpen()
    env.step(50)
    robot.IKTargetDoMove(position=[0, 0.5, 0], duration=2, speed_based=False, relative=True)
    env.step()
    robot.WaitDo()
    robot.IKTargetDoMove(position=[0, 0.5, 0.5], duration=2, speed_based=False)
    env.step()
    robot.WaitDo()
    box1.Destroy()
    box2.Destroy()
    env.step()



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr
from pyrfuniverse.side_channel import (
    IncomingMessage,
    OutgoingMessage,
)

env = RFUniverseBaseEnv(
    assets=['CustomAttr']
)

# custom message
custom = env.InstanceObject(name='CustomAttr', id=123456, attr_type=attr.CustomAttr)
custom.CustomMessage(message='this is instance channel custom message')
env.step()
print(custom.data['custom_message'])


# dynamic message
def dynamic_message_callback(msg: IncomingMessage):
    print(msg.read_string())
    print(msg.read_string())
    print(msg.read_string())
    print(msg.read_int32())
    print(msg.read_string())
    print(msg.read_float32())
    print(msg.read_string())
    print(msg.read_bool())
    print(msg.read_string())
    print(msg.read_float32_list())


env.AddListener('DynamicMessage', dynamic_message_callback)
env.SendMessage(
    'DynamicMessage',
    'string:',
    'this is dynamic message',
    'int:',
    123456,
    'bool:',
    True,
    'float:',
    4849.6564,
    'list:',
    [616445.085, 9489984.0, 65419596.0, 9849849.0]
)
env.step()


# dynamic object
def dynamic_object_callback(args):
    print(args[0])
    print(args[1])
    print(args[2])
    print(args[3])
    print(args[4])
    print(args[5])
    print(args[6])
    print(args[7])
    print(args[8])
    print(args[9])
    print(args[10])
    print(args[11])
    print(args[12])
    print(args[13])


env.AddListenerObject('DynamicObject', dynamic_object_callback)
env.SendObject(
    'DynamicObject',
    'string:',
    'this is dynamic object',
    'int:',
    123456,
    'bool:',
    True,
    'float:',
    4849.6564,
    'list:',
    [616445.085, 9489984.0, 65419596.0, 9849849.0],
    'dict:',
    {'1': 1, '2': 2, '3': 3},
    'tuple:',
    ('1', 1, 0.562)
)
env.step()

env.Pend()
env.close()



import pyrfuniverse.attributes as attr


class FallingClothAttr(attr.BaseAttr):
    def parse_message(self, data: dict):
        super().parse_message(data)

    def SetForceZoneParameters(self,
                               orientation: float = None,
                               intensity: float = None,
                               turbulence: float = None,
                               turbulence_frequency: float = None):
        self._send_data('SetForceZoneParameters',
                        orientation is not None, orientation,
                        intensity is not None, intensity,
                        turbulence is not None, turbulence,
                        turbulence_frequency is not None, turbulence_frequency)

    def SetSolverParameters(self, gravity: list):
        self._send_data('SetSolverParameters', gravity)



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.utils.rfuniverse_utility as utility

env = RFUniverseBaseEnv(
    scene_file='ArticulationIK.json'
)
ids = [639787, 985196, 221584, 8547820, 8547821]

for id in ids:
    current_robot = env.GetAttr(id)
    current_robot.IKTargetDoMove(position=[0, 0, -0.5], duration=0.1, relative=True)
    env.step()
    while not current_robot.data['move_done']:
        env.step()
    current_robot.IKTargetDoMove(position=[0, -0.5, 0], duration=0.1, relative=True)
    env.step()
    while not current_robot.data['move_done']:
        env.step()
    current_robot.IKTargetDoMove(position=[0, 0.5, 0.5], duration=0.1, relative=True)
    current_robot.IKTargetDoRotateQuaternion(quaternion=utility.UnityEularToQuaternion([90, 0, 0]), duration=30, relative=True)
    env.step()
    while not current_robot.data['move_done'] or not current_robot.data['rotate_done']:
        env.step()

env.Pend()
env.close()


from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
try:
    import gym
except ImportError:
    print('This feature requires gym, please install with `pip install gym==0.21.0`')
    raise


class RFUniverseGymGoalWrapper(gym.GoalEnv, RFUniverseBaseEnv):

    def __init__(
            self,
            executable_file: str = None,
            scene_file: str = None,
            assets: list = [],
            **kwargs

    ):
        RFUniverseBaseEnv.__init__(
            self,
            executable_file=executable_file,
            scene_file=scene_file,
            assets=assets,
            **kwargs,
        )

    def reset(self):
        gym.GoalEnv.reset(self)

    def close(self):
        RFUniverseBaseEnv.close(self)


from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.utils.rfuniverse_utility as utility
import os

env = RFUniverseBaseEnv()

ur5 = env.LoadURDF(path=os.path.abspath('../URDF/UR5/ur5_robot.urdf'), native_ik=True)
ur5.SetTransform(position=[1, 0, 0])
yumi = env.LoadURDF(path=os.path.abspath('../URDF/yumi_description/urdf/yumi.urdf'), native_ik=False)
yumi.SetTransform(position=[2, 0, 0])
kinova = env.LoadURDF(path=os.path.abspath('../URDF/kinova_gen3/GEN3_URDF_V12.urdf'), native_ik=False)
kinova.SetTransform(position=[3, 0, 0])
env.step()

ur5.IKTargetDoMove(position=[0, 0.5, 0], duration=0.1, relative=True)
env.step()
ur5.WaitDo()
ur5.IKTargetDoMove(position=[0, 0, -0.5], duration=0.1, relative=True)
env.step()
ur5.WaitDo()
ur5.IKTargetDoMove(position=[0, -0.2, 0.3], duration=0.1, relative=True)
ur5.IKTargetDoRotateQuaternion(quaternion=utility.UnityEularToQuaternion([0, 90, 0]), duration=30, relative=True)
env.step()
ur5.WaitDo()

env.Pend()



import os
import time
if os.name == "nt":
    import msvcrt

    def lock(fp):
        fp.seek(0)
        msvcrt.locking(fp.fileno(), msvcrt.LK_LOCK, 1)

    def unlock(fp):
        fp.seek(0)
        msvcrt.locking(fp.fileno(), msvcrt.LK_UNLCK, 1)

else:
    import fcntl

    def lock(fp):
        fcntl.flock(fp.fileno(), fcntl.LOCK_EX)

    def unlock(fp):
        fcntl.flock(fp.fileno(), fcntl.LOCK_UN)


class Locker:
    def __init__(self, lck_name: str):
        self.lck_path = os.path.join(os.path.expanduser("~"), ".rfuniverse", lck_name+'.lck')
        # check if the lock file exists
        if not os.path.exists(self.lck_path):
            os.makedirs(os.path.dirname(self.lck_path), exist_ok=True)
            open(self.lck_path, "w+").close()
        self.fp = None
    
    def __enter__(self):
        self.fp = open(self.lck_path)
        lock(self.fp)

    def __exit__(self, _type, value, tb):
        time.sleep(0.5)
        unlock(self.fp)
        self.fp.close()


import pyrfuniverse
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
from pyrfuniverse.utils.tobor_controller import RFUniverseToborController
import numpy as np


class ToborRobotiq85ManipulationEnv(RFUniverseBaseEnv):
    def __init__(
            self,
            task_name,
            scene_file=None,
            only_calculate=False,
            left_init_joint_positions=[0] * 7,
            right_init_joint_positions=[0] * 7
    ):
        super().__init__(
            scene_file=scene_file,
        )
        self.GetAttr(9874610).EnabledNativeIK(enabled=False)
        self.GetAttr(9874611).EnabledNativeIK(enabled=False)
        self._step()
        self.ik_controller = RFUniverseToborController(
            urdf_folder='../URDF/tobor',
            left_hand='robotiq85',
            right_hand='robotiq85',
            left_init_joint_positions=left_init_joint_positions,
            right_init_joint_positions=right_init_joint_positions,
            revise=True
        )

        # Init joint positions
        self.left_init_joint_positions = left_init_joint_positions
        self.left_joint_positions = self.left_init_joint_positions.copy()
        self.right_init_joint_positions = right_init_joint_positions
        self.right_joint_positions = self.right_init_joint_positions.copy()

        self.left_gripper_open = True
        self.right_gripper_open = True
        self.task_name = task_name
        self.record_file = '/home/haoyuan/workspace/rfuniverse/rfuniverse/RFUniverse/tobor_manipulation_{}.txt'.format(self.task_name)
        self.only_calculate = only_calculate
        #with open(self.record_file, 'w') as f:
            #f.close()

    def step(self, mode, position: np.ndarray, orientation=None):
        if mode == 'left':
            # Robotiq85
            current_position = np.array(self.GetAttr(98746100).data['positions'][7])
        else:
            # Robotiq85
            current_position = np.array(self.GetAttr(98746110).data['positions'][7])

        distance = position - current_position
        time_steps = int(np.abs(distance / 0.05).max()) + 1
        unit_distance = distance / time_steps

        for i in range(time_steps):
            target_position = current_position + unit_distance * (i + 1)
            joint_positions = self.ik_controller.calculate_ik(
                mode,
                target_position,
                orientation
            )

            if mode == 'left':
                self.left_joint_positions = joint_positions
                if not self.only_calculate:
                    self.GetAttr(9874610).SetJointPosition(
                        joint_positions=list(joint_positions),
                    )
            else:
                self.right_joint_positions = joint_positions
                if not self.only_calculate:
                    self.GetAttr(9874611).SetJointPosition(
                        joint_positions=list(joint_positions),
                    )

            for j in range(20):
                if not self.only_calculate:
                    self._step()
                self.write()


    def double_step(self, left_pos, right_pos, left_orn=None, right_orn=None):
        left_current_pos = np.array(self.GetAttr(98746100).data['positions'][7])
        left_distance = left_pos - left_current_pos
        left_time_steps = int(np.abs(left_distance / 0.05).max()) + 1

        right_current_pos = np.array(self.GetAttr(98746110).data['positions'][7])
        right_distance = right_pos - right_current_pos
        right_time_steps = int(np.abs(right_distance / 0.05).max()) + 1

        time_steps = max(left_time_steps, right_time_steps)
        left_unit_distance = left_distance / time_steps
        right_unit_distance = right_distance / time_steps

        for i in range(time_steps):
            left_target_position = left_current_pos + left_unit_distance * (i + 1)
            right_target_position = right_current_pos + right_unit_distance * (i + 1)
            left_joint_positions = self.ik_controller.calculate_ik(
                'left',
                left_target_position,
                left_orn
            )
            self.left_joint_positions = left_joint_positions
            right_joint_positions = self.ik_controller.calculate_ik(
                'right',
                right_target_position,
                right_orn
            )
            self.right_joint_positions = right_joint_positions

            if not self.only_calculate:
                self.GetAttr(9874610).SetJointPosition(
                    joint_positions=list(left_joint_positions),
                )
                self._step()
                self.GetAttr(9874611).SetJointPosition(
                    joint_positions=list(right_joint_positions),
                )

            for j in range(20):
                if not self.only_calculate:
                    self._step()
                self.write()

    def double_close(self):
        if not self.only_calculate:
            self.GetAttr(98746100).SetJointPosition(
                joint_positions=[50, 50]
            )
            self.left_gripper_open = False
            self._step()
            self.GetAttr(98746110).SetJointPosition(
                joint_positions=[50, 50]
            )
            self.right_gripper_open = False
            self._step()

        for i in range(20):
            if not self.only_calculate:
                self._step()
            self.write()

    def double_open(self):
        if not self.only_calculate:
            self.GetAttr(98746100).SetJointPosition(
                joint_positions=[0, 0]
            )
            self.left_gripper_open = True
            self._step()
            self.GetAttr(98746110).SetJointPosition(
                joint_positions=[0, 0]
            )
            self.right_gripper_open = True
            self._step()

        for i in range(20):
            if not self.only_calculate:
                self._step()
            self.write()

    def reset(self):
        self.GetAttr(9874610).SetJointPositionDirectly(
            joint_positions=self.left_init_joint_positions
        )
        self._step()
        self.GetAttr(9874611).SetJointPositionDirectly(
            joint_positions=self.right_init_joint_positions
        )
        self._step()

    def close_gripper(self, mode):
        if not self.only_calculate:
            if mode == 'left':
                self.GetAttr(98746100).SetJointPosition(
                    joint_positions=[50, 50]
                )
                self.left_gripper_open = False
            else:
                self.GetAttr(98746110).SetJointPosition(
                    joint_positions=[50, 50]
                )
                self.right_gripper_open = False
        for i in range(20):
            if not self.only_calculate:
                self._step()
            self.write()


    def open_gripper(self, mode):
        if not self.only_calculate:
            if mode == 'left':
                self.GetAttr(98746100).SetJointPosition(
                    joint_positions=[0, 0]
                )
                self.left_gripper_open = True
            else:
                self.GetAttr(98746110).SetJointPosition(
                    joint_positions=[0, 0]
                )
                self.right_gripper_open = True
        for i in range(20):
            if not self.only_calculate:
                self._step()
            self.write()


    def wait(self, n_timesteps):
        for i in range(n_timesteps):
            if not self.only_calculate:
                self._step()
            self.write()


    def write(self):
        '''
        with open(self.record_file, 'a+') as f:
            line = ''
            for joint_position in self.left_joint_positions:
                line += str(joint_position) + ','
            for joint_position in self.right_joint_positions:
                line += str(joint_position) + ','

            if self.left_gripper_open:
                line += '0,0,'
            else:
                line += '50,50,'

            if self.right_gripper_open:
                line += '0,0\n'
            else:
                line += '50,50\n'

            f.write(line)
            '''

    def write_raw(self, raw_str):
        with open(self.record_file, 'a+') as f:
            f.write(raw_str)

    def get_current_position(self, mode):
        if mode == 'left':
            return np.array(self.GetAttr(98746100).data['positions'][7])
        elif mode == 'right':
            return np.array(self.GetAttr(98746110).data['positions'][7])




from pyrfuniverse.attributes.base_attr import BaseAttr
from pyrfuniverse.attributes.camera_attr import CameraAttr
from pyrfuniverse.attributes.gameobject_attr import GameObjectAttr
from pyrfuniverse.attributes.light_attr import LightAttr
from pyrfuniverse.attributes.activelightsensor_attr import ActiveLightSensorAttr
from pyrfuniverse.attributes.collider_attr import ColliderAttr
from pyrfuniverse.attributes.controller_attr import ControllerAttr
from pyrfuniverse.attributes.rigidbody_attr import RigidbodyAttr
from pyrfuniverse.attributes.cloth_attr import ClothAttr
from pyrfuniverse.attributes.softbody_attr import SoftbodyAttr
from pyrfuniverse.attributes.humanbody_attr import HumanbodyAttr
from pyrfuniverse.attributes.graspsim_attr import GraspSimAttr
from pyrfuniverse.attributes.digit_attr import DigitAttr
from pyrfuniverse.attributes.pointcloud_attr import PointCloudAttr
from pyrfuniverse.attributes.fallingcloth_attr import FallingClothAttr
from pyrfuniverse.attributes.custom_attr import CustomAttr
try:
    from pyrfuniverse.attributes.omplmanager_attr import OmplManagerAttr
except ImportError:
    pass

__all__ = [
    'base_attr', 'camera_attr', 'activelightsensor_attr',
    'collider_attr', 'controller_attr', 'gameobject_attr',
    'rigidbody_attr', 'cloth_attr', 'softbody_attr',
    'humanbody_attr', 'graspsim_attr', 'digit_attr',
    'pointcloud_attr', 'fallingcloth_attr', 'custom_attr',
    'omplmanager_attr'
]



import os
import numpy as np
try:
    import pandas as pd
except ImportError:
    raise Exception('This feature requires pandas, please install with `pip install pandas`')
try:
    import open3d as o3d
except ImportError:
    raise Exception('This feature requires open3d, please install with `pip install open3d`')
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr


def get_grasp_pose(file: str, points_count, scale: float = 1):
    mesh = o3d.io.read_triangle_mesh(file)
    mesh.scale(scale, np.array([0, 0, 0]))
    sample_points = mesh.sample_points_poisson_disk(points_count, use_triangle_normal=True)
    sample_points.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
    points = np.asarray(sample_points.points)
    normals = np.asarray(sample_points.normals)
    return points, normals


mesh_path = '../Mesh/drink1/drink1.obj'

points, normals = get_grasp_pose(mesh_path, 100)
points = points.reshape(-1).tolist()
normals = normals.reshape(-1).tolist()

env = RFUniverseBaseEnv(assets=['GraspSim'])
grasp_sim = env.InstanceObject(id=123123, name='GraspSim', attr_type=attr.GraspSimAttr)
grasp_sim.StartGraspSim(mesh=os.path.abspath(mesh_path),
                        gripper='franka_hand',
                        points=points,
                        normals=normals,
                        depth_range_min=-0.05,
                        depth_range_max=0,
                        depth_lerp_count=5,
                        angle_lerp_count=5,
                        parallel_count=100
                        )

# only show grasp pose
# grasp_sim.GenerateGraspPose(mesh=os.path.abspath(mesh_path),
#                             gripper='SimpleFrankaGripper',
#                             points=points,
#                             normals=normals,
#                             depth_range_min=-0.05,
#                             depth_range_max=0,
#                             depth_lerp_count=5,
#                             angle_lerp_count=5,
#                             )
env.step()
while not grasp_sim.data['done']:
    env.step()

points = grasp_sim.data['points']
points = np.array(points).reshape([-1, 3])
quaternions = grasp_sim.data['quaternions']
quaternions = np.array(quaternions).reshape([-1, 4])
width = grasp_sim.data['width']
width = np.array(width).reshape([-1, 1])

env.close()

data = np.concatenate((points, quaternions, width), axis=1)
csv = pd.DataFrame(data, columns=['x', 'y', 'z', 'qx', 'qy', 'qz', 'qw', 'width'])

csv_path = os.path.join(os.path.dirname(mesh_path), 'grasps_rfu.csv')
csv.to_csv(csv_path, index=True, header=True)

env.Pend()
env.close()



from pyrfuniverse.envs import RFUniverseGymWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding


class BalanceBallEnv(RFUniverseGymWrapper):
    metadata = {'render.modes': ['human']}

    def __init__(self, executable_file=None):
        super().__init__(
            executable_file,
            rigidbody_channel=True,
            game_object_channel=True,
        )
        self.action_space = spaces.Box(low=-2.0, high=2.0, shape=(2,), dtype=np.float32)
        self.observation_space = spaces.Box(
            low=np.array([-1, -1, -3, -2, -3, -float('inf'), -float('inf'), -float('inf')]),
            high=np.array([1, 1, 3, 4, 3, float('inf'), float('inf'), float('inf')]),
            dtype=np.float32
        )
        self.t = 0
        self.r = 0
        self.max_steps = 50

    def step(self, a: np.ndarray):
        """
        Params:
            a: 2-d numpy array. The first dimension is for cube's x axis rotation, while the second dimension is for
            cube's z axis rotation.
        """
        self.game_object_channel.set_action(
            'Rotate',
            index=0,
            rotation=[a[0], 0, a[1]]
        )
        self._step()

        done = False
        info = {}
        cube_position = self.game_object_channel.data[0]['position']
        sphere_position = self.rigidbody_channel.data[0]['position']
        is_fail = self._check_fail(cube_position, sphere_position)
        if is_fail:
            self.r = -1
            done = True
            self.reset()
            info['done'] = True
            info['is_success'] = False
        elif self.t == self.max_steps:
            self.r = 0.1
            done = True
            self.reset()
            info['done'] = True
            info['is_success'] = True
        else:
            self.r = 0.1
            self.t += 1
            info['done'] = False

        return self._get_obs(), self.r, done, info

    def _get_obs(self):
        cube_quaternion = self.game_object_channel.data[0]['quaternion']
        cube_position = self.game_object_channel.data[0]['position']
        sphere_position = self.rigidbody_channel.data[0]['position']
        sphere_velocify = self.rigidbody_channel.data[0]['velocity']

        rotation = np.array([cube_quaternion[0], cube_quaternion[2]])
        relative_position = np.array(sphere_position) - np.array(cube_position)
        sphere_velocify = np.array(sphere_velocify)

        return np.concatenate((rotation, relative_position, sphere_velocify))

    def reset(self):
        self.t = 0

        cubeRotationX = self._generate_random_float(-10, 10)
        cubeRotationZ = self._generate_random_float(-10, 10)
        spherePositionX = self._generate_random_float(-1.5, 1.5)
        spherePositionZ = self._generate_random_float(-1.5, 1.5)

        self.env_param_channel.set_float_parameter('cubeRotationX', cubeRotationX)
        self.env_param_channel.set_float_parameter('cubeRotationZ', cubeRotationZ)
        self.env_param_channel.set_float_parameter('spherePositionX', spherePositionX)
        self.env_param_channel.set_float_parameter('spherePositionZ', spherePositionZ)

        self.env.reset()
        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode='human'):
        self._step()

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, \
            'Min value is {}, while max value is {}.'.format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _check_fail(self, cube_position, sphere_position):
        if sphere_position[1] - cube_position[1] < -2:
            return True

        if abs(sphere_position[0] - cube_position[0]) > 3:
            return True

        if abs(sphere_position[2] - cube_position[2]) > 3:
            return True

        return False


class BalanceBallEnvV0(BalanceBallEnv):

    def __init__(self):
        super().__init__('/home/haoyuan/workspace/rfuniverse/build/BalanceBall/RFUniverse.x86_64')



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

env = RFUniverseBaseEnv(assets=['tobor_r300_ag95_ag95'])

torbor = env.InstanceObject(name='tobor_r300_ag95_ag95', attr_type=attr.ControllerAttr)
torbor.SetTransform(position=[0, 0.05, 0])
torbor.SetImmovable(False)
env.step()
while 1:
    torbor.MoveForward(1, 0.2)
    env.step(300)
    torbor.TurnLeft(90, 30)
    env.step(300)
    torbor.MoveForward(1, 0.2)
    env.step(300)
    torbor.TurnLeft(90, 30)
    env.step(300)
    torbor.MoveForward(1, 0.2)
    env.step(300)
    torbor.TurnRight(90, 30)
    env.step(300)
    torbor.MoveBack(1, 0.2)
    env.step(300)
    torbor.TurnRight(90, 30)
    env.step(300)




import pybullet as p
import pybullet_data
import numpy as np
import math


ll = [-7] * 7
# upper limits for null space (todo: set them to proper range)
ul = [7] * 7
# joint ranges for null space (todo: set them to proper range)
jr = [7] * 7
# restposes for null space
# jointPositions = [0, -1 * math.pi / 4, 0, -3 * math.pi / 4, 0, math.pi / 2, math.pi / 4, 0.04, 0.04]
jointPositions = [0.00, 0.41, 0.00, -1.85, -0.00, 2.26, 0.79, 0.04, 0.04]
rp = jointPositions

class RFUniverseController:
    """
    RFUniverseController is a class to generate robot arm joint states. In simulation environment, we mostly
    want to specify the 6DoF of a joint, then the robot arm will automatically move to that state. Thus, here
    we use pybullet.calculateInverseKinematics() to generate joint positions based on a given robot arm, a
    given end-effector joint and a target Cartesian position. The generated joint states will be passed to
    Unity by rfuniverse channels. Besides, this class will also provide functions to align coordinate in Unity
    and in pybullet.

    TODO: Currently, this class has many hard codes for Franka, need further refinement.
    """

    def __init__(
            self, robot_name, robot_urdf=None, base_pos=np.array([0, 0, 0]),
            base_orn=[-0.707107, 0.0, 0.0, 0.707107], end_effector_id=None,
            render=False
    ):
        if render:
            p.connect(p.GUI) # For debug mode
        else:
            p.connect(p.DIRECT)

        p.configureDebugVisualizer(p.COV_ENABLE_Y_AXIS_UP, 1)
        p.setAdditionalSearchPath(pybullet_data.getDataPath())
        p.setGravity(0, -9.8, 0)

        self.bullet_client = p
        self.bullet_flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES

        self.robot_name = robot_name
        self.robot_urdf = robot_urdf
        self.end_effector_id = end_effector_id
        self.num_dof = None
        self._parse_robot_params()

        self.robot = self.bullet_client.loadURDF(self.robot_urdf, base_pos, base_orn, useFixedBase=True,
                                                 flags=self.bullet_flags)

        # Magic starts
        self.revolute_joint_ids = []
        self.prismatic_joint_ids = []
        index = 0
        for j in range(self.bullet_client.getNumJoints(self.robot)):
            self.bullet_client.changeDynamics(self.robot, j, linearDamping=0, angularDamping=0)
            info = self.bullet_client.getJointInfo(self.robot, j)

            jointName = info[1]
            jointType = info[2]
            if (jointType == self.bullet_client.JOINT_PRISMATIC):
                self.bullet_client.resetJointState(self.robot, j, jointPositions[index])
                self.prismatic_joint_ids.append(j)
                index = index + 1
            if (jointType == self.bullet_client.JOINT_REVOLUTE):
                self.bullet_client.resetJointState(self.robot, j, jointPositions[index])
                self.revolute_joint_ids.append(j)
                index = index + 1

        # print('pybullet', self.get_link_state(self.end_effector_id))

    def _parse_robot_params(self):
        if self.robot_name == 'franka':
            self.robot_urdf = 'franka_panda/panda.urdf'
            self.end_effector_id = 11
            self.num_dof = 7
            return
        elif self.robot_name == 'tobor':
            self.end_effector_id = 6
            self.num_dof = 7
            return
        elif self.robot_name == 'ur5':
            self.num_dof = 6

        # elif self.robot_name == 'another_known_robot':
        #     self.robot_urdf = 'known_urdf_path'
        #     return

        elif self.robot_urdf is None:
            print('Error: an unknown robot name without an input urdf.')
            exit(-1)

    def get_bullet_pos_from_unity(self, unity_pos: list) -> list:
        return [-1 * unity_pos[0], unity_pos[1], unity_pos[2]]

    def get_unity_pos_from_bullet(self, bullet_pos: list) -> list:
        return [-1 * bullet_pos[0], bullet_pos[1], bullet_pos[2]]

    def get_unity_joint_pos_from_pybullet(self, pybullet_joint_pos: tuple) -> list:
        pybullet_joint_pos = list(pybullet_joint_pos)[:self.num_dof]
        for i, (joint_pos) in enumerate(pybullet_joint_pos):
            pybullet_joint_pos[i] = -180 * joint_pos / math.pi

        return pybullet_joint_pos

    def calculate_ik(self, unity_eef_pos, eef_orn=None) -> list:
        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)

        joint_positions = self.bullet_client.calculateInverseKinematics(
            self.robot, self.end_effector_id, eef_pos, eef_orn, ll, ul, jr, rp, maxNumIterations=20)

        for i, (idx) in enumerate(self.revolute_joint_ids):
            self.bullet_client.resetJointState(self.robot, idx, joint_positions[i])

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def calculate_ik_recursive(self, unity_eef_pos, eef_orn=None) -> list:
        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)
        for i in range(20):
            joint_positions = self.bullet_client.calculateInverseKinematics(
                self.robot, self.end_effector_id, eef_pos, eef_orn, ll, ul, jr, rp, maxNumIterations=20)

            for i, (idx) in enumerate(self.revolute_joint_ids):
                self.bullet_client.resetJointState(self.robot, idx, joint_positions[i])

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def get_link_state(self, link_idx):
        link_state = self.bullet_client.getLinkState(self.robot, link_idx)

        return self.get_unity_pos_from_bullet(link_state[0])

    def reset(self):
        for i, (idx) in enumerate(self.revolute_joint_ids):
            self.bullet_client.resetJointState(self.robot, idx, jointPositions[i])



import threading
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr
import cv2
import numpy as np

img = None
class ImageThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)

    def run(self):
        while True:
            global img
            if img is not None:
                cv2.imshow('image', img)
                cv2.waitKey(10)

env = RFUniverseBaseEnv(assets=['Camera', 'GameObject_Box'])

camera = env.InstanceObject(name='Camera', id=123456, attr_type=attr.CameraAttr)
camera.SetTransform(position=[0, 0.25, 0], rotation=[30, 0, 0])
box = env.InstanceObject(name='GameObject_Box', attr_type=attr.GameObjectAttr)
box.SetTransform(position=[0, 0.05, 0.5], scale=[0.1, 0.1, 0.1])
box.SetColor([1, 0, 0, 1])

thread = ImageThread()
thread.start()
while True:
    camera.GetRGB(width=512, height=512)
    box.Rotate([0, 1, 0], False)
    env.step()
    image = np.frombuffer(camera.data['rgb'], dtype=np.uint8)
    img = cv2.imdecode(image, cv2.IMREAD_COLOR)



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
from pyrfuniverse.attributes.light_attr import LightType

env = RFUniverseBaseEnv(scene_file='LightScene.json')

light = env.GetAttr(885275)
env.SetShadowDistance(50)
while 1:
    env.step(50)
    light.SetColor(color=[1., 0., 0.])
    env.step(50)
    light.SetRange(30.)
    env.step(50)
    light.SetType(LightType.Directional)
    env.step(50)
    light.SetIntensity(5.)
    env.step(50)
    light.SetType(LightType.Spot)
    env.step(50)
    light.SetSpotAngle(60.)
    env.step(50)
    light.SetType(LightType.Point)
    env.step(50)
    light.SetRange(10.)
    light.SetIntensity(1.)
    light.SetSpotAngle(30.)



import random
import subprocess
from abc import ABC
import socket
import numpy as np
import pyrfuniverse
import pyrfuniverse.attributes as attr
from pyrfuniverse.side_channel import IncomingMessage, OutgoingMessage
from pyrfuniverse.utils.locker import Locker
from pyrfuniverse.utils.rfuniverse_communicator import RFUniverseCommunicator
import os


class RFUniverseBaseEnv(ABC):
    """
    RFUniverse base environment class.

    Args:
        executable_file: Str, the absolute path of Unity executable file. None for last used executable file; "@editor" for using Unity Editor.
        scene_file: Str, the absolute path of Unity scene JSON file. All JSON files locate at `StraemingAssets/SceneData` by default.
        assets: List, the list of pre-load assets. All assets in the list will be pre-loaded in Unity when the environment is initialized, which will save time during instanciating.
        graphics: Bool, True for showing GUI and False for headless mode.
    """

    metadata = {"render.modes": ["human", "rgb_array"]}

    def __init__(
        self,
        executable_file: str = None,
        scene_file: str = None,
        assets: list = [],
        graphics: bool = True,
        port: int = 5004,
        proc_id=0,
        log_level=1,
    ):
        # time step
        self.t = 0
        self.executable_file = executable_file
        self.scene_file = scene_file
        self.pre_load_assets = assets
        self.graphics = graphics
        self.process = None
        self.attrs = {}
        self.data = {}
        self.listen_messages = {}
        self.listen_object = {}
        self.port = port

        self.log_level = log_level

        self.log_map = {"Log": 3, "Warning": 2, "Error": 1, "Exception": 1, "Assert": 1}

        if self.executable_file is None:
            self.executable_file = pyrfuniverse.executable_file

        PROC_TYPE = ""  # editor or release
        if self.executable_file == "" or self.executable_file == "@editor":  # editor
            assert proc_id == 0, "proc_id must be 0 when using editor"
            print("Waiting for UnityEditor play...")
            PROC_TYPE = "editor"
            self.port = 5004
        elif os.path.exists(self.executable_file):  # release
            PROC_TYPE = "release"
            self.port = 5005 + proc_id  # default release port
        else:  # error
            raise ValueError(f"Executable file {self.executable_file} not exists")

        self.communicator = RFUniverseCommunicator(
            port=self.port,
            receive_data_callback=self._receive_data,
            proc_type=PROC_TYPE,
        )
        self.port = self.communicator.port  # update port
        if PROC_TYPE == "release":
            with Locker('config'): # unity process will try to modify the config file
                self.process = self._start_unity_env(self.executable_file, self.port)
        self.communicator.online()
        self._send_debug_data("SetPythonVersion", pyrfuniverse.__version__)
        if len(self.pre_load_assets) > 0:
            self.PreLoadAssetsAsync(assets, True)
        if self.scene_file is not None:
            self.LoadSceneAsync(self.scene_file, True)

    def _get_port(self) -> int:
        executable_port = 5005
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while True:
            try:
                s.bind(("localhost", executable_port))
                s.close()
                return executable_port
            except OSError:
                executable_port += 1
                continue

    def _start_unity_env(self, executable_file: str, port: int) -> subprocess.Popen:
        arg = [executable_file]
        if not self.graphics:
            arg.extend(["-nographics", "-batchmode"])
        if self.log_level == 0:
            proc_out = subprocess.DEVNULL
        else:
            proc_out = None
        arg.append(f"-port:{port}")
        return subprocess.Popen(arg, stdout=proc_out, stderr=proc_out)

    def _receive_data(self, objs: list) -> None:
        msg = objs[0]
        objs = objs[1:]
        if msg == "Env":
            self._parse_env_data(objs)
        elif msg == "Instance":
            self._parse_instence_data(objs)
        elif msg == "Debug":
            self._parse_debug_data(objs)
        elif msg == "Message":
            self._parse_message_data(objs)
        elif msg == "Object":
            self._parse_object_data(objs)
        return

    def _parse_env_data(self, objs: list) -> None:
        msg = objs[0]
        objs = objs[1:]
        if msg == "Close":
            self.close()
        elif msg == "LoadDone":
            self.data["load_done"] = True
        elif msg == "PendDone":
            self.data["pend_done"] = True
        elif msg == "RFMoveColliders":
            self.data["colliders"] = objs[0]
        elif msg == "CurrentCollisionPairs":
            self.data["collision_pairs"] = objs[0]
        else:
            print(f"unknown env data type: {msg}")

    def _parse_instence_data(self, objs: list) -> None:
        this_object_id = objs[0]
        this_object_type = objs[1]
        this_object_data = objs[2]

        attr_type = eval("attr." + this_object_type)
        if this_object_id not in self.attrs:
            self.attrs[this_object_id] = attr_type(self, this_object_id)
        elif type(self.attrs[this_object_id]) != attr_type:
            self.attrs[this_object_id] = attr_type(
                self, this_object_id, self.attrs[this_object_id].data
            )

        self.data[this_object_id] = self.attrs[this_object_id].parse_message(
            this_object_data
        )

    def _parse_debug_data(self, objs: list) -> None:
        msg = objs[0]
        objs = objs[1:]
        if msg == "Log":
            if self.log_map[objs[0]] <= self.log_level:
                print(
                    f"Unity Env Log Type:{objs[0]}\nCondition:{objs[1]}\nStackTrace:{objs[2]}"
                )
        else:
            print(f"unknown debug data type: {msg}")
        return

    def _parse_message_data(self, objs: list) -> None:
        msg = objs[0]
        objs = objs[1:]
        if msg in self.listen_messages:
            self.listen_messages[msg](IncomingMessage(objs[0]))

    def _parse_object_data(self, objs: list) -> None:
        head = objs[0]
        objs = objs[1:]
        if head in self.listen_object:
            self.listen_object[head](objs)

    def _send_env_data(self, *args) -> None:
        self.communicator.send_object("Env", *args)

    def _send_instance_data(self, *args) -> None:
        self.communicator.send_object("Instance", *args)

    def _send_debug_data(self, *args) -> None:
        self.communicator.send_object("Debug", *args)

    def _send_message_data(self, *args) -> None:
        self.communicator.send_object("Message", *args)

    def _send_object_data(self, *args) -> None:
        self.communicator.send_object("Object", *args)

    def _step(self):
        """
        Send the messages of called functions to Unity and simulate for a step, then accept the data from Unity.
        """
        if not self.communicator.connected:
            raise Exception("Unity Env not connected")
        self.communicator.sync_step()

    def step(self, count: int = 1):
        """
        Send the messages of called functions to Unity and simulate for a step, then accept the data from Unity.

        Args:
            count: the number of steps for executing Unity simulation.
        """
        if count < 1:
            count = 1
        for _ in range(count):
            self._step()

    def close(self):
        """
        Close the environment
        """
        if self.process is not None:
            self.process.kill()
        if self.communicator is not None:
            self.communicator.close()

    def GetAttr(self, id: int):
        """
        Get the attribute instance by object id.

        Args:
            id: Int, object id.

        Returns:
            pyrfuniverse.attributes.BaseAttr: An instance of attribute.
        """
        assert id in self.attrs, "this ID not exists"
        return self.attrs[id]

    # Env API
    def PreLoadAssetsAsync(self, names: list, auto_wait: bool = False) -> None:
        """
        PreLoad the asset.

        Args:
            names: list, the name of assets.
            auto_wait: Bool, if True, this function will not return until the loading is done.
        """
        self._send_env_data("PreLoadAssetsAsync", names)

        if auto_wait:
            self.WaitLoadDone()

    def LoadSceneAsync(self, file: str, auto_wait: bool = False) -> None:
        """
        Load the scene asynchronisely.

        Args:
            file: Str, the scene JSON file. If it's a relative path, it will load from `StraemingAssets`.
            auto_wait: Bool, if True, this function will not return until the loading is done.
        """
        self._send_env_data("LoadSceneAsync", file)

        if auto_wait:
            self.WaitLoadDone()

    def SwitchSceneAsync(self, name: str, auto_wait: bool = False) -> None:
        """
        Switch the scene asynchronisely.

        Args:
            name: Str, the scene name.
            auto_wait: Bool, if True, this function will not return until the loading is done.
        """
        self._send_env_data("SwitchSceneAsync", name)

        if auto_wait:
            self.WaitLoadDone()

    def WaitLoadDone(self) -> None:
        """
        Wait for the loading is done.
        """
        self.data["load_done"] = False
        while not self.data["load_done"]:
            self._step()

    def Pend(self) -> None:
        """
        Pend the program until the `EndPend` button in `UnityPlayer` is clicked.
        """
        self._send_env_data("Pend")

        self.data["pend_done"] = False
        while not self.data["pend_done"]:
            self._step()

    def SendMessage(self, message: str, *args) -> None:
        """
        Send message to Unity.

        Args:
            message: Str, the message head.
            *args: List, the list of parameters. We support str, bool, int, float and List[float] types.
        """
        msg = OutgoingMessage()
        for i in args:
            if type(i) == str:
                msg.write_string(i)
            elif type(i) == bool:
                msg.write_bool(i)
            elif type(i) == int:
                msg.write_int32(i)
            elif type(i) == float or type(i) == np.float32 or type(i) == np.float64:
                msg.write_float32(float(i))
            elif type(i) == list and type(i[0]) == float:
                msg.write_float32_list(i)
            else:
                print(f"dont support this data type:{type(i)}")
        self._send_message_data(message, msg.buffer)

    def SendObject(self, head: str, *args) -> None:
        """
        Send object to Unity.

        Args:
            head: Str, the message head.
            *args: List, the list of parameters. We support str, bool, int, float and List[float] types.
        """
        self._send_object_data(head, *args)

    def AddListener(self, message: str, fun):
        """
        Add listener.

        Args:
            message: Str, the message head.
            fun: Callable, the callback function.
        """
        self.listen_messages[message] = fun

    def RemoveListener(self, message: str, fun):
        """
        Remove listener.

        Args:
            message: Str, the message head.
            fun: Callable, the callback function.
        """
        self.listen_messages.pop(message)

    def AddListenerObject(self, type: str, fun):
        """
        Add object listener.

        Args:
            type: Str, the message head.
            fun: Callable, the callback function.
        """
        self.listen_object[type] = fun

    def RemoveListenerObject(self, type: str):
        """
        Remove object listener.

        Args:
            type: Str, the message head.
        """
        self.listen_object.pop(type)

    def InstanceObject(
        self, name: str, id: int = None, attr_type: type = attr.BaseAttr
    ):
        """
        Instanciate an object.

        Built-in assets:

        GameObjectAttr:
            Basic Objects:
                "GameObject_Box",
                "GameObject_Capsule",
                "GameObject_Cylinder",
                "GameObject_Sphere",
                "GameObject_Quad",
            IGbison Meshes:
                "Hainesburg_mesh_texture",
                "Halfway_mesh_texture",
                "Hallettsville_mesh_texture",
                "Hambleton_mesh_texture",
                "Hammon_mesh_texture",
                "Hatfield_mesh_texture",
                "Haxtun_mesh_texture",
                "Haymarket_mesh_texture",
                "Hendrix_mesh_texture",
                "Hercules_mesh_texture",
                "Highspire_mesh_texture",
                "Hitchland_mesh_texture",

        ColliderAttr:
            "Collider_Box",
            "Collider_ObiBox",
            "Collider_Capsule",
            "Collider_Cylinder",
            "Collider_Sphere",
            "Collider_Quad",

        RigidbodyAttr:
            Basic Objects:
                "Rigidbody_Box",
                "GameObject_Capsule",
                "Rigidbody_Cylinder",
                "Rigidbody_Sphere",
            YCB dataset:
                77 models in YCB dataset. See YCB Object and Model Set for detail: https://rse-lab.cs.washington.edu/projects/posecnn/

        ControllerAttr:
            gripper:
                "allegro_hand_right",
                "bhand",
                "svh",
                "robotiq_arg2f_85_model",
                "dh_robotics_ag95_gripper",
            robot:
                "kinova_gen3",
                "kinova_gen3_robotiq85",
                "ur5",
                "ur5_robotiq85",
                "franka_panda",
                "franka_hand",
                "tobor_robotiq85_robotiq85",
                "flexivArm",
                "flexivArm_ag95",
                "yumi",

        CameraAttr:
            "Camera",

        LightAttr:
            "Light",

        Args:
            name: Str, object name. Please check the above `built-in assets` list for names.
            id: Int, object id.
            attr_type: type(pyrfuniverse.attributes.BaseAttr), the attribute type.

        Returns:
            type(`attr_type`): The object attribute instance.
        """
        assert id not in self.attrs, "this ID exists"

        while id is None or id in self.attrs:
            id = random.randint(100000, 999999)

        self._send_env_data("InstanceObject", name, id)

        self.attrs[id] = attr_type(self, id)
        return self.attrs[id]

    def LoadURDF(
        self, path: str, id: int = None, native_ik: bool = False, axis: str = "y"
    ) -> attr.ControllerAttr:
        """
        Load a model from URDF file.

        Args:
            path: Str, the URDF file path.
            id: Int, object id.
            native_ik: Bool, True for enabling native IK; False for using custom IK.When it is True, through the IKTargetDo*** interface, according to the end pose.When it is False, through the SetJoint*** interface, according to the joint movement.
            axis: Str, import axis, This can be 'z' or 'y', depending on the URDF file

        Returns:
            pyrfuniverse.attributes.ControllerAttr: The object attribute intance.
        """
        assert id not in self.attrs, "this ID exists"

        while id is None or id in self.attrs:
            id = random.randint(100000, 999999)

        self._send_env_data("LoadURDF", id, path, native_ik, axis)

        self.attrs[id] = attr.ControllerAttr(self, id)
        return self.attrs[id]

    def LoadMesh(self, path: str, id: int = None) -> attr.RigidbodyAttr:
        """
        Load a model from Mesh file.

        Args:
            path: Str, the Mesh file path.
            id: Int, object id.

        Returns:
            pyrfuniverse.attributes.RigidbodyAttr: The object attribute intance.
        """
        assert id not in self.attrs, "this ID exists"

        while id is None or id in self.attrs:
            id = random.randint(100000, 999999)

        self._send_env_data("LoadMesh", id, path)

        self.attrs[id] = attr.RigidbodyAttr(self, id)
        return self.attrs[id]

    def IgnoreLayerCollision(self, layer1: int, layer2: int, ignore: bool) -> None:
        """
        Ignore or enable the collision between two layers.

        Args:
            layer1: Int, the layer number of the first layer.
            layer2: Int, the layer number of the second layer.
            ignore: Bool, True for ignoring collision between two layers; False for enabling collision between two layers.
        """
        self._send_env_data("IgnoreLayerCollision", layer1, layer2, ignore)

    def GetCurrentCollisionPairs(self) -> None:
        """
        Get the collision pairs of current collision. After calling this method and stepping once, the result will be saved in env.data['CurrentCollisionPairs']
        """
        self._send_env_data("GetCurrentCollisionPairs")

    def GetRFMoveColliders(self) -> None:
        """
        Get the RFMove colliders. After calling this method and stepping once, the result will be saved in env.data['RFMoveColliders']
        """
        self._send_env_data("GetRFMoveColliders")

    def SetGravity(self, x: float, y: float, z: float) -> None:
        """
        Set the gravity of environment.

        Args:
            x: Float, gravity on global x-axis (right).
            y: Float, gravity on global y-axis (up).
            z: Float, gravity on global z-axis (forward).
        """
        self._send_env_data("SetGravity", [float(x), float(y), float(z)])

    def SetGroundActive(self, active: bool) -> None:
        """
        Set the ground active or inactive.

        Args:
            active: Bool, active or inactive the ground.
        """
        self._send_env_data("GetRFMoveColliders", active)

    def SetGroundPhysicMaterial(
        self,
        bounciness: float,
        dynamic_friction: float,
        static_friction: float,
        friction_combine: int,
        bounce_combine: int,
    ) -> None:
        """
        Set the physics material of ground in environment.

        Args:
            bounciness: Float, the bounciness.
            dynamic_friction: Float, the dynamic friction coefficient (0-1).
            static_friction: Float, the static friction coefficient (0-1).
            friction_combine: Int, how friction of two colliding objects is combined. 0 for Average, 1 for Minimum, 2 for Maximum and 3 for Multiply. See https://docs.unity3d.com/Manual/class-PhysicMaterial.html for more details.
            bounce_combine: Int, how bounciness of two colliding objects is combined. The value representation is the same with `friction_combine`.
        """
        self._send_env_data(
            "SetGroundPhysicMaterial",
            float(bounciness),
            float(dynamic_friction),
            float(static_friction),
            friction_combine,
            bounce_combine,
        )

    def SetTimeStep(self, delta_time: float) -> None:
        """
        Set the time for a step in Unity.

        Args:
            delta_time: Float, the time for a step in Unity.
        """
        self._send_env_data("SetTimeStep", float(delta_time))

    def SetTimeScale(self, time_scale: float) -> None:
        """
        Set the time scale in Unity.

        Args:
            time_scale: Float, the time scale in Unity.
        """
        self._send_env_data("SetTimeScale", float(time_scale))

    def SetResolution(self, resolution_x: int, resolution_y: int) -> None:
        """
        Set the resolution of windowed GUI.

        Args:
            resolution_x: Int, window width.
            resolution_y: Int, window height.
        """
        self._send_env_data("SetResolution", resolution_x, resolution_y)

    def ExportOBJ(self, items_id: list, save_path: str) -> None:
        """
        Export the specified object list to OBJ file. For native bundle models, the `Read/Write` must be checked in Unity Editor.

        Args:
            items_id: List, the object ids.
            save_path: Str, the path to save the OBJ files.
        """
        self._send_env_data("ExportOBJ", items_id, save_path)

    def SetShadowDistance(self, distance: float) -> None:
        """
        Set the shadow distance for rendering in environment.

        Args:
            distance: Float, the shadow distance measured in meter.
        """
        self._send_env_data("SetShadowDistance", float(distance))

    def SaveScene(self, file: str) -> None:
        """
        Save current scene.

        Args:
            file: Str, the file path to save current scene. Default saving to `StreamingAssets` folder.
        """
        self._send_env_data("SaveScene", file)

    def ClearScene(self) -> None:
        """
        Clear current scene.
        """
        self._send_env_data("ClearScene")

    def AlignCamera(self, camera_id: int) -> None:
        """
        Align current GUI view to a given camera.

        Args:
            camera_id: Int, camera id.
        """
        self._send_env_data("AlignCamera", camera_id)

    def SetViewTransform(self, position: list = None, rotation: list = None) -> None:
        """
        Set the GUI view.

        Args:
            position: A list of length 3, representing the position of GUI view.
            rotation: A list of length 3, representing the rotation of GUI view.
        """
        if position is not None:
            assert (
                type(position) == list and len(position) == 3
            ), "Argument position must be a 3-d list."
            position = [float(i) for i in position]
        if rotation is not None:
            assert (
                type(rotation) == list and len(rotation) == 3
            ), "Argument rotation must be a 3-d list."
            rotation = [float(i) for i in rotation]

        self._send_env_data("SetViewTransform", position, rotation)

    def ViewLookAt(self, target: list, world_up: list = None) -> None:
        """
        Rotates the transform so the forward vector points at target's current position.

        Args:
            target: A list of length 3, target to point towards.
            world_up: A list of length 3, vector specifying the upward direction.
        """
        if world_up is None:
            world_up = [0., 1., 0.]
        assert len(target) == 3, 'target length must be 3'
        target = [float(i) for i in target]
        assert len(world_up) == 3, 'world_up length must be 3'
        world_up = [float(i) for i in world_up]

        self._send_env_data('ViewLookAt', target, world_up)

    def SetViewBackGround(self, color: list = None) -> None:
        """
        Set the GUI view.

        Args:
            color: A list of length 3, background color of GUI view. None : default skybox.
        """
        if color is not None:
            assert type(color) == list and len(color) == 3, "color length must be 3"
            color = [float(i) for i in color]

        self._send_env_data("SetViewBackGround", color)

    # Dubug API
    def DebugGraspPoint(self, enabled: bool = True) -> None:
        """
        Show or hide end effector of robot arm for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("DebugGraspPoint", enabled)

    def DebugObjectPose(self, enabled: bool = True) -> None:
        """
        Show or hide object base point for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("DebugObjectPose", enabled)

    def DebugCollisionPair(self, enabled: bool = True) -> None:
        """
        Show or hide collision pairs for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("DebugCollisionPair", enabled)

    def DebugColliderBound(self, enabled: bool = True) -> None:
        """
        Show or hide collider bounding box for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("DebugColliderBound", enabled)

    def DebugObjectID(self, enabled: bool = True) -> None:
        """
        Show or hide object id for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("DebugObjectID", enabled)

    def Debug3DBBox(self, enabled: bool = True) -> None:
        """
        Show or hide 3d bounding box of objects for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("Debug3DBBox", enabled)

    def Debug2DBBox(self, enabled: bool = True) -> None:
        """
        Show or hide 2d bounding box of objects for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("Debug2DBBox", enabled)

    def DebugJointLink(self, enabled: bool = True) -> None:
        """
        Show or hide joint information of articulation for debug.

        Args:
            enabled: Bool, True for showing and False for hiding.
        """
        self._send_debug_data("DebugJointLink", enabled)

    def SendLog(self, log: str) -> None:
        """
        Send log messange and show it on Unity GUI window.

        Args:
            log: Str, log message.
        """
        self._send_debug_data("SendLog", log)

    def ShowArticulationParameter(self, controller_id: int) -> None:
        """
        Show Articulation Parameter on Unity GUI window.

        Args:
            controller_id: int, controller_attr id.
        """
        self._send_debug_data("ShowArticulationParameter", int(controller_id))



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import os

env = RFUniverseBaseEnv()
kinova = env.LoadURDF(path=os.path.abspath('../URDF/Franka/panda.urdf'), axis='z')
kinova.SetTransform(position=[0, 0, 0])
env.ShowArticulationParameter(kinova.id)

env.Pend()



import pybullet as p
import pybullet_data
import numpy as np
import math
import os


class RFUniverseToborController:

    def __init__(
            self,
            urdf_folder,
            left_base_pos=np.array([0.07, 1, 0.22]),
            left_base_orn=p.getQuaternionFromEuler([-math.pi / 2, math.pi, -math.pi / 2]),
            left_hand='robotiq85',
            right_base_pos=np.array([-0.07, 1, 0.22]),
            right_base_orn=p.getQuaternionFromEuler([-math.pi / 2, 0, math.pi / 2]),
            right_hand='dhag95',
            ik_tolerance=0.03,
            render=False,
            left_init_joint_positions=[0] * 7,
            right_init_joint_positions=[0] * 7,
            revise=False
    ):
        self.revise = revise
        if render:
            p.connect(p.GUI)
        else:
            p.connect(p.DIRECT)
        p.configureDebugVisualizer(p.COV_ENABLE_Y_AXIS_UP, 1)
        p.setAdditionalSearchPath(urdf_folder)
        p.setGravity(0, -9.8, 0)
        self.bullet_client = p
        self.bullet_flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES

        self.urdf_folder = urdf_folder
        self.left_base_pos = left_base_pos
        self.left_base_orn = left_base_orn
        self.left_hand_name = left_hand
        self.left_urdf = 'tobor_arm_' + self.left_hand_name + '.urdf'

        self.right_base_pos = right_base_pos
        self.right_base_orn = right_base_orn
        self.right_hand_name = right_hand
        self.right_urdf = 'tobor_arm_' + self.right_hand_name + '.urdf'

        self._load_urdf()
        self.ik_tolerance = ik_tolerance

        # end_effector_id: 7 robot arm joints plus 1 fixed joint for grasp point
        self.end_effector_id = 7
        self.num_dof = 7
        self.revolute_joint_ids = list(range(7))
        self.link_ids = list(range(8))

        # revise_factor: fix the positive or negative joint position between pybullet and Unity.
        self.revise_factor_old = np.array([-1, 1, -1, 1, -1, 1, -1])
        self.revise_factor = np.array([1, 1, 1, 1, 1, 1, 1])
        self.init_joint_positions = {
            'left': self.get_pybullet_joint_pos_from_unity(left_init_joint_positions),
            'right': self.get_pybullet_joint_pos_from_unity(right_init_joint_positions),
        }

        self.reset()

        # For debug
        joint_info = self.bullet_client.getJointInfo(self.left_arm_id, 0)
        print(joint_info[1], joint_info[12], joint_info[16])

    def get_bullet_pos_from_unity(self, unity_pos: list) -> list:
        return [-1 * unity_pos[0], unity_pos[1], unity_pos[2]]

    def get_unity_pos_from_bullet(self, bullet_pos: list) -> list:
        return [-1 * bullet_pos[0], bullet_pos[1], bullet_pos[2]]

    def get_unity_joint_pos_from_pybullet(self, pybullet_joint_pos: tuple) -> list:
        pybullet_joint_pos = list(pybullet_joint_pos)[:self.num_dof]
        pybullet_joint_pos = np.array(pybullet_joint_pos)
        if self.revise is True:
            unity_joint_pos = self.revise_factor_old * pybullet_joint_pos * 180 / math.pi
        else:
            unity_joint_pos = self.revise_factor * pybullet_joint_pos * 180 / math.pi

        return unity_joint_pos

    def get_pybullet_joint_pos_from_unity(self, unity_joint_pos: list) -> list:
        unity_joint_pos = list(unity_joint_pos)[:self.num_dof]
        unity_joint_pos = np.array(unity_joint_pos)
        if self.revise is True:
            pybullet_joint_pos = self.revise_factor_old * unity_joint_pos * math.pi / 180
        else:
            pybullet_joint_pos = self.revise_factor * unity_joint_pos * math.pi / 180

        return pybullet_joint_pos

    def calculate_ik(self, mode, unity_eef_pos, eef_orn=None) -> list:
        self._check_mode(mode)
        if mode == 'left':
            robot_id = self.left_arm_id
        else:
            robot_id = self.right_arm_id

        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)

        for j in range(10):
            joint_positions = self.bullet_client.calculateInverseKinematics(
                robot_id, self.end_effector_id, eef_pos, eef_orn, maxNumIterations=20)

            for i, (idx) in enumerate(self.revolute_joint_ids):
                self.bullet_client.resetJointState(robot_id, idx, joint_positions[i])

            link_state = self.bullet_client.getLinkState(robot_id, self.end_effector_id)
            distance = self._calculate_distance(link_state[4], eef_pos)
            if distance < self.ik_tolerance:
                break

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def get_link_position(self, mode, link_id):
        """
        Return (link_world_position, world_link_frame_position)
        link_world_position: The Cartesian position of center of mass.
        world_link_frame_position: The position of URDF link frame.
        """
        self._check_mode(mode)
        if mode == 'left':
            link_state = self.bullet_client.getLinkState(self.left_arm_id, link_id)
        else:
            link_state = self.bullet_client.getLinkState(self.right_arm_id, link_id)

        return (self.get_unity_pos_from_bullet(link_state[0]),
                self.get_unity_pos_from_bullet(link_state[4]))

    def get_all_link_positions(self, mode):
        self._check_mode(mode)
        positions = []

        for link_id in self.link_ids:
            link_world_position, world_link_frame_position = self.get_link_position(mode, link_id)
            positions.append(world_link_frame_position)

        return positions

    def reset(self):
        for i, (idx) in enumerate(self.revolute_joint_ids):
            self.bullet_client.resetJointState(self.left_arm_id, idx, self.init_joint_positions['left'][i])
            self.bullet_client.resetJointState(self.right_arm_id, idx, self.init_joint_positions['right'][i])

    def _load_urdf(self):
        self.right_arm_id = self.bullet_client.loadURDF(
            self.right_urdf,
            self.right_base_pos,
            self.right_base_orn,
            useFixedBase=True,
            flags=self.bullet_flags
        )

        self.left_arm_id = self.bullet_client.loadURDF(
            self.left_urdf,
            self.left_base_pos,
            self.left_base_orn,
            useFixedBase=True,
            flags=self.bullet_flags
        )

    def _check_mode(self, mode):
        assert mode in ['left', 'right'], \
            'Error: Calculating IK must work in mode \'left\' or \'right\'.'

    def _calculate_distance(self, pos1, pos2):
        pos1 = np.array(pos1)
        pos2 = np.array(pos2)
        distance = np.linalg.norm(pos1 - pos2)

        return distance



import math
import numpy as np

def EncodeIDAsColor(instance_id: int):
    """
    Encode the object id to a color.

    Args:
        instance_id: Int, the id of the object.
    
    Return:
        List: The encoded color in [r, g, b, 255] format.
    """
    r = (instance_id * 16807 + 187) % 256
    g = (instance_id * 48271 + 79) % 256
    b = (instance_id * 95849 + 233) % 256
    return [r, g, b, 255]

def UnityEularToQuaternion(eular: list) -> list:
    """
    Transform euler angle to quaternion in Unity.

    Args:
        eular: List of length 3, representing euler angle in [x, y, z]
        order and measured in degree.

    Return:
        List: The transformed quaternion in [x, y, z, w] format.
    """
    xx = math.radians(eular[0])
    yy = math.radians(eular[1])
    zz = math.radians(eular[2])
    x = math.cos(yy / 2) * math.sin(xx / 2) * math.cos(zz / 2) + math.sin(yy / 2) * math.cos(xx / 2) * math.sin(zz / 2)
    y = math.sin(yy / 2) * math.cos(xx / 2) * math.cos(zz / 2) - math.cos(yy / 2) * math.sin(xx / 2) * math.sin(zz / 2)
    z = math.cos(yy / 2) * math.cos(xx / 2) * math.sin(zz / 2) - math.sin(yy / 2) * math.sin(xx / 2) * math.cos(zz / 2)
    w = math.cos(yy / 2) * math.cos(xx / 2) * math.cos(zz / 2) + math.sin(yy / 2) * math.sin(xx / 2) * math.sin(zz / 2)
    return [x, y, z, w]

def UnityQuaternionToEular(quaternion: list) -> list:
    """
    Transform quaternion to euler angle in Unity.

    Args:
        quaternion: List of length 4, representing quaternion in [x, y, z, w]
        order.

    Return:
        List: The transformed euler angle in [x, y, z] order and measured in degree.
    """
    xx = quaternion[0]
    yy = quaternion[1]
    zz = quaternion[2]
    ww = quaternion[3]
    x = math.asin(2 * ww * xx - 2 * yy * zz)
    y = math.atan2(2 * ww * yy + 2 * xx * zz, 1 - 2 * xx * xx - 2 * yy * yy)
    z = math.atan2(2 * ww * zz + 2 * xx * yy, 1 - 2 * xx * xx - 2 * zz * zz)
    x = math.degrees(x)
    y = math.degrees(y)
    z = math.degrees(z)
    return [x, y, z]

def CheckKwargs(kwargs: dict, compulsory_params: list):
    legal = True
    for param in compulsory_params:
        if param not in kwargs.keys():
            legal = False
            assert legal, \
                'Parameters illegal, parameter <%s> missing.' % param

def GetMatrix(pos, quat) -> np.ndarray:
    """
    Transform the position and quaternion into a transformation matrix.

    Args:
        pos: List of length 3, representing the [x, y, z] position.
        quat: List of length 4, representing the [x, y, z, w] quaternion.

    Return:
        numpy.ndarray: the transformation matrix.
    """
    q = quat.copy()
    n = np.dot(q, q)
    if n < np.finfo(q.dtype).eps:
        return np.identity(4)
    q = q * np.sqrt(2.0 / n)
    q = np.outer(q, q)
    # rot_matrix = np.array(
    #     [[1.0 - q[2, 2] - q[3, 3], q[1, 2] + q[3, 0], q[1, 3] - q[2, 0], pos[0]],
    #      [q[1, 2] - q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] + q[1, 0], pos[1]],
    #      [q[1, 3] + q[2, 0], q[2, 3] - q[1, 0], 1.0 - q[1, 1] - q[2, 2], pos[2]],
    #      [0, 0, 0, 1.0]], dtype=q.dtype)
    matrix = np.array(
        [[1.0 - q[1, 1] - q[2, 2], -(q[2, 3] - q[1, 0]), q[1, 3] + q[2, 0], pos[0]],
         [q[2, 3] + q[1, 0], -(1.0 - q[1, 1] - q[3, 3]), q[1, 2] - q[3, 0], pos[1]],
         [-(q[1, 3] - q[2, 0]), q[1, 2] + q[3, 0], -(1.0 - q[2, 2] - q[3, 3]), pos[2]],
         [0., 0., 0., 1.0]], dtype=float)
    return matrix




from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper
import numpy as np
import math
import copy
from gym import spaces
from gym.utils import seeding


class FrankaClothEnv(RFUniverseGymGoalWrapper):
    metadata = {'render.modes': ['human']}

    def __init__(
            self,
            asset_bundle_file,
            reward_type,
            tolerance=0.08,
            target_x_range=(-0.1, -0.1),
            target_z_range=(-0.1, 0.1),
            executable_file=None
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
            game_object_channel=True,
            obi_cloth_channel=True,
        )
        self.asset_bundle_file = asset_bundle_file
        self.reward_type = reward_type
        self.tolerance = tolerance
        self.target_x_range = target_x_range
        self.target_z_range = target_z_range

        # Fixed parameters
        self.obi_solver_position = np.array([-0.1, 0.31, -0.1])
        self.number_of_particles = 289
        self.number_of_sampled_particles = 10

        # Env setup
        self.load_cloth = False
        self._load_obi_cloth()
        self.ik_controller = RFUniverseController('franka', base_pos=np.array([-0.6, 0, 0]))
        self.eef_orn = self.ik_controller.bullet_client.getQuaternionFromEuler([math.pi / 2, 0, math.pi / 2])

        self.t = 0
        self.goal = self._sample_goal()
        self.action_space = spaces.Box(
            low=-1, high=1, shape=(4,), dtype=np.float32
        )
        obs = self._get_obs()
        self.observation_space = spaces.Dict({
            'observation': spaces.Box(-np.inf, np.inf, shape=obs['observation'].shape, dtype=np.float32),
            'desired_goal': spaces.Box(-np.inf, np.inf, shape=obs['desired_goal'].shape, dtype=np.float32),
            'achieved_goal': spaces.Box(-np.inf, np.inf, shape=obs['achieved_goal'].shape, dtype=np.float32)
        })

    def step(self, action: np.ndarray):
        pos_ctrl = action[:3] * 0.05
        curr_pos = np.array(self.articulation_channel.data[1]['positions'][3])
        pos_ctrl = curr_pos + pos_ctrl
        joint_positions = self.ik_controller.calculate_ik(
            pos_ctrl,
            eef_orn=self.eef_orn
        )

        curr_gripper_width = self._get_gripper_width()
        target_gripper_width = curr_gripper_width + action[3] * 0.2
        target_gripper_width = np.clip(target_gripper_width, 0, 0.08)
        joint_positions.append(target_gripper_width)

        self._set_franka_joints(np.array(joint_positions))
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {
            'is_success': self._check_success(obs)
        }
        reward = self.compute_reward(obs['achieved_goal'], obs['desired_goal'], info)

        return obs, reward, done, info

    def reset(self):
        super().reset()

        self._destroy_obi_cloth()
        self.env.reset()
        self._load_obi_cloth()
        self.goal = self._sample_goal()

        self.t = 0
        self.ik_controller.reset()
        joint_positions = self.ik_controller.calculate_ik(
            [0, 0.3, 0.1],
            self.eef_orn
        )
        self.articulation_channel.set_action(
            'SetJointPositionDirectly',
            index=0,
            joint_positions=joint_positions,
        )

        self.game_object_channel.set_action(
            'SetTransform',
            index=0,
            position=list(self.goal.copy())
        )

        self._step()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode='human'):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        distance = self._compute_goal_distance(achieved_goal, desired_goal)
        if self.reward_type == 'sparse':
            return -(distance > self.tolerance).astype(np.float32)
        else:
            return -distance

    def _get_obs(self):
        gripper_position = np.array(self.articulation_channel.data[1]['positions'][3])
        gripper_velocity = np.array(self.articulation_channel.data[1]['velocities'][3])
        gripper_width = self._get_gripper_width()
        panda_obs = np.concatenate((gripper_position, gripper_velocity, [gripper_width]))

        cloth_particles = np.array(self.obi_cloth_channel.data[0]['positions'])
        random_indices = np.random.randint(0, self.number_of_particles, size=(self.number_of_sampled_particles,))
        sampled_particles = cloth_particles[random_indices] + self.obi_solver_position
        sampled_particles = np.reshape(sampled_particles, newshape=(3 * self.number_of_sampled_particles,))

        obs = np.concatenate((panda_obs, sampled_particles))

        cloth_center = np.average(cloth_particles, axis=0) + self.obi_solver_position

        return {
            'observation': obs.copy(),
            'achieved_goal': cloth_center.copy(),
            'desired_goal': self.goal.copy()
        }

    def _load_obi_cloth(self):
        if self.load_cloth:
            return
        self.asset_channel.set_action(
            'LoadObiCloth',
            filename=self.asset_bundle_file,
            name='Obi Solver',
            position=list(self.obi_solver_position)
        )
        self._step()
        self.load_cloth = True

    def _destroy_obi_cloth(self):
        if not self.load_cloth:
            return
        self.obi_cloth_channel.set_action(
            'Destroy',
            index=0
        )
        self._step()
        self.load_cloth = False

    def _get_gripper_width(self):
        gripper_joint_positions = copy.deepcopy(self.articulation_channel.data[1]['joint_positions'])
        return -1 * (gripper_joint_positions[0] + gripper_joint_positions[1])

    def _set_franka_joints(self, a: np.ndarray):
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=0,
            joint_positions=list(a[0:7]),
        )
        self._step()

        a[7] = -1 * a[7] / 2
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=1,
            joint_positions=[a[7], a[7]],
        )
        self._step()

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min <= max, \
            'Min value is {}, while max value is {}.'.format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _compute_goal_distance(self, goal_a, goal_b):
        assert goal_a.shape == goal_b.shape
        return np.linalg.norm(goal_a - goal_b, axis=-1)

    def _check_success(self, obs):
        achieved_goal = obs['achieved_goal']
        desired_goal = obs['desired_goal']
        distance = self._compute_goal_distance(achieved_goal, desired_goal)

        return (distance < self.tolerance).astype(np.float32)

    def _sample_goal(self):
        target_x = self._generate_random_float(self.target_x_range[0], self.target_x_range[1])
        target_z = self._generate_random_float(self.target_z_range[0], self.target_z_range[1])

        return np.array([target_x, 0.3, target_z])



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr
import cv2
import numpy as np

env = RFUniverseBaseEnv(assets=['Camera', 'Rigidbody_Sphere_Draggable'])

camera = env.InstanceObject(name='Camera', attr_type=attr.CameraAttr)
camera.SetTransform(position=[-0.1, 0.033, 0.014], rotation=[0, 90, 0])
target = env.InstanceObject(name='Rigidbody_Sphere_Draggable')
target.SetTransform(position=[0, 0.05, 0.015])
env.step()
env.AlignCamera(camera.id)
env.SendLog("click End Pend button to start heat map record")
env.Pend()
camera.StartHeatMapRecord([target.id])
env.SendLog("click End Pend button to end heat map record")
env.Pend()
camera.EndHeatMapRecord()
camera.GetHeatMap()
env.step()
print(camera.data['heat_map'])
image_np = np.frombuffer(camera.data['heat_map'], dtype=np.uint8)
image_np = cv2.imdecode(image_np, cv2.IMREAD_COLOR)
print(image_np.shape)
env.close()
cv2.imshow("heatmap", image_np)
cv2.waitKey(0)



from pyrfuniverse.envs.gripper_nail.nail_card_env import NailCardEnv
from pyrfuniverse.envs.gripper_nail.nail_can_env import NailCanEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_can_env import Robotiq85NailCanEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_card_env import Robotiq85NailCardEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_book_env import Robotiq85NailBookEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_nail_coin_env import Robotiq85NailCoinEnv
from pyrfuniverse.envs.gripper_nail.robotiq85_insert_env import Robotiq85InsertEnv

__all__ = [
    'NailCardEnv', 'NailCanEnv', 'Robotiq85NailCanEnv', 'Robotiq85NailCardEnv', 'Robotiq85NailBookEnv',
    'Robotiq85NailCoinEnv', 'Robotiq85InsertEnv'
]


import os
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

env = RFUniverseBaseEnv(scene_file='SimpleYCBModel.json')

model = []

for i in env.attrs:
    if type(env.attrs[i]) is attr.RigidbodyAttr:
        model.append(i)

env.ExportOBJ(model, os.path.abspath('../Mesh/scene_mesh.obj'))

env.Pend()
env.close()




import random
import os
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

env = RFUniverseBaseEnv()
env.step()
mesh = env.LoadMesh(path=os.path.abspath('../Mesh/002_master_chef_can/google_16k/textured.obj'))
mesh.SetTransform(position=[0, 1, 0], rotation=[random.random() * 360, random.random() * 360, random.random() * 360])

for i in range(100):
    env.step(20)
    new_mesh = mesh.Copy(new_id=mesh.id + i + 1)
    new_mesh.SetTransform(position=[0, 1, 0], rotation=[random.random() * 360, random.random() * 360, random.random() * 360])

while 1:
    env.step()



from pyrfuniverse.envs.multi_physics.kinova_gen2_catching_cloth_env import KinovaGen2CatchingClothEnv
from pyrfuniverse.envs.multi_physics.ur5_water_shooting import UR5WaterShootingEnv
from pyrfuniverse.envs.multi_physics.flexiv_cutting import FlexivCuttingEnv

__all__ = [
    'KinovaGen2CatchingClothEnv', 'UR5WaterShootingEnv', 'FlexivCuttingEnv'
]

try:
    from gym.envs.registration import register

    register(
        id='KinovaCatchingCloth-v4',
        entry_point='pyrfuniverse.envs.multi_physics:KinovaGen2CatchingClothEnv',
        kwargs={
            'lock_eef_height': True,
            'with_force_zone': True,
            'max_steps': 100,
            # 'executable_file': join_path('KinovaGen2CatchingCloth/RFUniverse.x86_64')
        },
    )

except ImportError:
    print('No gym installed. Please install gym!')
    pass



import os.path
import pyrfuniverse.attributes as attr
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

env = RFUniverseBaseEnv()
env.SetViewBackGround([0.,0.,0.])
point_cloud = env.InstanceObject(name='PointCloud', id=123456, attr_type=attr.PointCloudAttr)
point_cloud.ShowPointCloud(ply_path=os.path.abspath('../Mesh/000000_000673513312.ply'))
point_cloud.SetTransform(rotation=[-90, 0, 0])
point_cloud.SetRadius(radius=0.001)

env.Pend()
env.close()




import os
try:
    import pandas as pd
except ImportError:
    print('This feature requires pandas, please install with `pip install pandas`')
    raise
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

mesh_path = '../Mesh/drink1/drink1.obj'
pose_path = '../Mesh/drink1/grasps_rfu.csv'

data = pd.read_csv(pose_path, usecols=['x', 'y', 'z', 'qx', 'qy', 'qz', 'qw'])
data = data.to_numpy()
positions = data[:, 0:3].reshape(-1).tolist()
quaternions = data[:, 3:7].reshape(-1).tolist()

env = RFUniverseBaseEnv()
grasp_sim = env.InstanceObject(id=123123, name='GraspSim', attr_type=attr.GraspSimAttr)
grasp_sim.ShowGraspPose(mesh=os.path.abspath(mesh_path), gripper='SimpleFrankaGripper', positions=positions, quaternions=quaternions)

env.Pend()
env.close()



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

env = RFUniverseBaseEnv(scene_file='DebugScene.json')

while 1:
    env.DebugGraspPoint()
    env.SendLog("DebugGraspPoint")
    env.step(300)
    env.DebugGraspPoint(False)

    env.DebugObjectID()
    env.SendLog("DebugObjectID")
    env.step(300)
    env.DebugObjectID(False)

    env.DebugObjectPose()
    env.SendLog("DebugObjectPose")
    env.step(300)
    env.DebugObjectPose(False)

    env.DebugColliderBound()
    env.SendLog("DebugColliderBound")
    env.step(300)
    env.DebugColliderBound(False)

    env.DebugCollisionPair()
    env.SendLog("DebugCollisionPair")
    env.step(300)
    env.DebugCollisionPair(False)

    env.Debug3DBBox()
    env.SendLog("Debug3DBBox")
    env.step(300)
    env.Debug3DBBox(False)

    env.Debug2DBBox()
    env.SendLog("Debug2DBBox")
    env.step(300)
    env.Debug2DBBox(False)

    env.DebugJointLink()
    env.SendLog("DebugJointLink")
    env.step(300)
    env.DebugJointLink(False)
